class DefaultType:
    def __init__(self, value, obj):
        self.value = value
        self.obj = obj

    def default(self):
        if callable(self.value):
            return self.value(self.obj)
        else:
            return self.value

    def __getattr__(self, attr):
        return self.default()

    def __call__(self, *args, **kwargs):
        return self.default()

    def __eq__(self, other):
        value = self.default()
        if value is None:
            return other is value
        else:
            return other == value


class Optional:
    """Optional helper class that allow evaluting an expression which can be undefined without
    raising an expression but returning a default value (None)."""

    def __init__(self, obj, default=None):
        self.obj = obj
        self.default = default

    def __getattr__(self, attr):
        if hasattr(self.obj, attr):
            return getattr(self.obj, attr)
        return DefaultType(self.default, self.obj)()

    def __call__(self, *args, **kwargs):
        return DefaultType(self.default, self.obj)()

    def instance(self):
        return self.obj



from .NotificationManager import NotificationManager
from .MockNotification import MockNotification
from .DatabaseNotification import DatabaseNotification
from .Notification import Notification
from .Notifiable import Notifiable
from .AnonymousNotifiable import AnonymousNotifiable
from .Sms import Sms
from .SlackMessage import SlackMessage
from .Textable import Textable



import os
import platform
import pathlib


def make_directory(directory):
    """Create a directory at the given path for a file if it does not exist"""
    if not os.path.isfile(directory):
        if not os.path.exists(os.path.dirname(directory)):
            # Create the path to the model if it does not exist
            os.makedirs(os.path.dirname(directory))

        return True

    return False


def file_exists(directory):
    """Create a directory at the given path for a file if it does not exist"""
    return os.path.exists(os.path.dirname(directory))


def make_full_directory(directory):
    """Create all directories to the given path if they do not exist"""
    if not os.path.isfile(directory):
        if not os.path.exists(directory):
            # Create the path to the model if it does not exist
            os.makedirs(directory)

        return True

    return False


def creation_date(path_to_file):
    """Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    """
    if platform.system() == "Windows":
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime


def modified_date(path_to_file):
    if platform.system() == "Windows":
        return os.path.getmtime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_mtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return 0


def render_stub_file(stub_file, name):
    """Read stub file, replace placeholders and return content."""
    with open(stub_file, "r") as f:
        content = f.read()
        content = content.replace("__class__", name)
    return content


def get_module_dir(module_file):
    return os.path.dirname(os.path.realpath(module_file))


def get_extension(filepath: str, without_dot=False) -> str:
    """Get file extension from a filepath. If without_dot=True the . prefix will be removed from
    the extension."""
    extension = "".join(pathlib.Path(filepath).suffixes)
    if without_dot:
        return extension[1:]
    return extension



import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
from ..Recipient import Recipient
import ssl


class SMTPDriver:
    def __init__(self, application):
        self.application = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def get_mime_message(self):
        message = MIMEMultipart("alternative")

        message["Subject"] = self.options.get("subject")

        message["From"] = Recipient(self.options.get("from")).header()
        message["To"] = Recipient(self.options.get("to")).header()
        if self.options.get("reply_to"):
            message["Reply-To"] = Recipient(self.options.get("reply_to")).header()

        if self.options.get("cc"):
            message["Cc"] = Recipient(self.options.get("cc")).header()

        if self.options.get("bcc"):
            message["Bcc"] = Recipient(self.options.get("bcc")).header()

        if self.options.get("html_content"):
            message.attach(MIMEText(self.options.get("html_content"), "html"))

        if self.options.get("text_content"):
            message.attach(MIMEText(self.options.get("text_content"), "plain"))

        if self.options.get("priority"):
            message["X-Priority"] = self.options.get("priority")

        if self.options.get("headers"):
            for header, value in self.options.get("headers").items():
                message[header] = value

        for attachment in self.options.get("attachments", []):
            with open(attachment.path, "rb") as fil:
                part = MIMEApplication(fil.read(), Name=attachment.alias)

            part["Content-Disposition"] = f"attachment; filename={attachment.alias}"
            message.attach(part)

        return message

    def make_connection(self):
        options = self.options
        if options.get("ssl"):
            smtp = smtplib.SMTP_SSL("{0}:{1}".format(options["host"], options["port"]))
        else:
            smtp = smtplib.SMTP("{0}:{1}".format(options["host"], int(options["port"])))

        if options.get("tls"):
            context = ssl.create_default_context()
            context.check_hostname = False

            # Check if correct response code for starttls is received from the server
            if smtp.starttls(context=context)[0] != 220:
                raise smtplib.SMTPNotSupportedError(
                    "Server is using untrusted protocol."
                )

        if options.get("username") and options.get("password"):
            smtp.login(options.get("username"), options.get("password"))

        return smtp

    def send(self):
        smtp = self.make_connection()
        smtp.send_message(self.get_mime_message())



from ..drivers.queue import DatabaseDriver, AsyncDriver, AMQPDriver
from ..queues import Queue
from ..configuration import config


class QueueProvider:
    def __init__(self, application):
        self.application = application

    def register(self):
        queue = Queue(self.application).set_configuration(config("queue.drivers"))
        queue.add_driver("database", DatabaseDriver(self.application))
        queue.add_driver("async", AsyncDriver(self.application))
        queue.add_driver("amqp", AMQPDriver(self.application))
        self.application.bind("queue", queue)

    def boot(self):
        pass



from cleo import CommandTester
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..foundation import Application


class TestCommand:
    """This class allows to test craft commands and asserts command outputs."""

    def __init__(self, application: "Application"):
        self.application = application

    def run(self, command: str, arguments_str: str = "") -> "TestCommand":
        """Run the given craft command. Commands arguments should be provided as
        one string (space separated)."""
        command = self.application.make("commands").command_application.find(command)
        self.command_tester = CommandTester(command)
        self.command_tester.execute(arguments_str)
        return self

    def assertExactOutput(self, ref_output: str) -> "TestCommand":
        """Assert command output to be exactly the same as the given string."""
        output = self._get_output()
        assert ref_output == output, f"Command output was: {output}, not {ref_output}"
        return self

    def assertOutputContains(self, ref_output: str) -> "TestCommand":
        """Assert command output to contain the given string."""
        output = self._get_output()
        assert (
            ref_output in output
        ), f"Command output was: {output} and does not contain {ref_output}."
        return self

    def assertOutputMissing(self, ref_output: str) -> "TestCommand":
        """Assert command output does not contain the given reference output."""
        output = self._get_output()
        assert (
            ref_output not in output
        ), f"Command output was: {output}, not {ref_output}"
        return self

    def assertHasErrors(self) -> "TestCommand":
        assert self._get_errors()
        return self

    def assertExactErrors(self, ref_errors: str) -> "TestCommand":
        errors = self._get_errors()
        assert (
            errors == ref_errors
        ), f"Command output has errors: {errors}, not {ref_errors}."
        return self

    def assertSuccess(self) -> "TestCommand":
        """Assert that command returned a 0 exit code meaning that it ran successfully."""
        code = self.command_tester.status_code
        assert 0 == code, "Command exited code is not 0: {code}."
        return self

    def _get_errors(self) -> str:
        return self.command_tester.io.fetch_error()

    def _get_output(self) -> str:
        return self.command_tester.io.fetch_output()



import pickle
import pendulum
import inspect
from urllib import parse

from ...utils.console import HasColoredOutput


class AMQPDriver(HasColoredOutput):
    def __init__(self, application):
        self.application = application
        self.connection = None
        self.publishing_channel = None

    def set_options(self, options):
        self.options = options
        return self

    def push(self, *jobs, args=(), **kwargs):
        for job in jobs:
            payload = {
                "obj": job,
                "args": args,
                "callback": self.options.get("callback", "handle"),
                "created": pendulum.now(tz=self.options.get("tz", "UTC")),
            }

            try:
                self.connect().publish(payload)
            except (self.get_connection_exceptions()):
                self.connect().publish(payload)

    def get_connection_exceptions(self):
        pika = self.get_package_library()
        return (
            pika.exceptions.ConnectionClosed,
            pika.exceptions.ChannelClosed,
            pika.exceptions.ConnectionWrongStateError,
            pika.exceptions.ChannelWrongStateError,
        )

    def publish(self, payload):
        pika = self.get_package_library()
        self.publishing_channel.basic_publish(
            exchange=self.options.get("exchange"),
            routing_key=self.options.get("queue"),
            body=pickle.dumps(payload),
            properties=pika.BasicProperties(
                delivery_mode=2,  # make message persistent
            ),
        )
        self.publishing_channel.close()
        self.connection.close()

    def get_package_library(self):
        try:
            import pika
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'pika' library. Run 'pip install pika' to fix this."
            )

        return pika

    def connect(self):
        try:
            import pika
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'pika' library. Run 'pip install pika' to fix this."
            )
        connection_url = "amqp://{}:{}@{}{}/{}".format(
            self.options.get("username"),
            self.options.get("password"),
            self.options.get("host"),
            ":" + str(self.options.get("port")) if self.options.get("port") else "",
            self.options.get("vhost", "%2F"),
        )
        if self.options.get("connection_options"):
            connection_url += "?" + parse.urlencode(
                self.options.get("connection_options")
            )
        self.connection = pika.BlockingConnection(pika.URLParameters(connection_url))

        self.publishing_channel = self.connection.channel()

        self.publishing_channel.queue_declare(self.options.get("queue"), durable=True)

        return self

    def consume(self):
        self.success(
            '[*] Waiting to process jobs on the "{}" queue. To exit press CTRL+C'.format(
                self.options.get("queue")
            )
        )

        self.connect()

        self.publishing_channel.basic_qos(prefetch_count=1)

        self.publishing_channel.basic_consume(self.options.get("queue"), self.work)

        try:
            self.publishing_channel.start_consuming()
        finally:
            self.publishing_channel.stop_consuming()
            self.publishing_channel.close()
            self.connection.close()

    def retry(self):
        builder = (
            self.application.make("builder")
            .new()
            .on(self.options.get("connection"))
            .table(self.options.get("failed_table", "failed_jobs"))
        )

        jobs = builder.get()

        if len(jobs) == 0:
            self.success("No failed jobs found.")
            return

        for job in jobs:
            try:
                self.connect().publish(pickle.loads(job["payload"]))
            except (self.get_connection_exceptions()):
                self.connect().publish(pickle.loads(job["payload"]))

        self.success(f"Added {len(jobs)} failed jobs back to the queue")
        builder.table(self.options.get("failed_table", "failed_jobs")).where_in(
            "id", [x["id"] for x in jobs]
        ).delete()

    def work(self, ch, method, _, body):

        job = pickle.loads(body)
        obj = job["obj"]
        args = job["args"]
        callback = job["callback"]

        try:
            try:
                if inspect.isclass(obj):
                    obj = self.application.resolve(obj)

                getattr(obj, callback)(*args)

            except AttributeError:
                obj(*args)

            self.success(
                f"[{method.delivery_tag}][{pendulum.now(tz=self.options.get('tz', 'UTC')).to_datetime_string()}] Job Successfully Processed"
            )
        except Exception as e:
            self.danger(
                f"[{method.delivery_tag}][{pendulum.now(tz=self.options.get('tz', 'UTC')).to_datetime_string()}] Job Failed"
            )

            getattr(obj, "failed")(job, str(e))

            self.add_to_failed_queue_table(
                self.application.make("builder").new(), str(job["obj"]), body, str(e)
            )

        ch.basic_ack(delivery_tag=method.delivery_tag)

    def add_to_failed_queue_table(self, builder, name, payload, exception):
        builder.table(self.options.get("failed_table", "failed_jobs")).create(
            {
                "driver": "amqp",
                "queue": self.options.get("queue", "default"),
                "name": name,
                "connection": self.options.get("connection"),
                "created_at": pendulum.now(
                    tz=self.options.get("tz", "UTC")
                ).to_datetime_string(),
                "exception": exception,
                "payload": payload,
                "failed_at": pendulum.now(
                    tz=self.options.get("tz", "UTC")
                ).to_datetime_string(),
            }
        )



"""Verify Email Module."""

import time

from ..auth.Sign import Sign


class MustVerifyEmail:
    """Class To Verify User Email."""

    def verify_email(self, mail_manager, request):
        """Sends email for user verification

        Arguments:
            mail_manager {masonite.managers.MailManager} -- Masonite mail manager class.
            request {masonite.request.Request} -- Masonite request class.
        """
        mail = mail_manager.helper()
        sign = Sign()

        token = sign.sign("{0}::{1}".format(self.id, time.time()))
        link = "{0}/email/verify/{1}".format(request.environ["HTTP_HOST"], token)

        mail.to(self.email).template(
            "auth/verifymail", {"name": self.name, "email": self.email, "link": link}
        ).subject("Please Confirm Your Email").send()



import pendulum
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from ..foundation import Application
    from .limiters import Limiter


class RateLimiter:
    def __init__(self, application: "Application"):
        self.application = application
        self.limiters: dict = {}

    def register(self, name, callback: "Limiter") -> "RateLimiter":
        self.limiters[name] = callback
        return self

    @property
    def cache(self) -> Any:
        """Get default cache driver"""
        return self.application.make("cache").store()

    def attempts(self, key: str) -> int:
        key = self.clean_key(key)
        return int(self.cache.get(key, default=0))

    def clean_key(self, key: str) -> str:
        """Clean the rate limiter key from unicode characters."""
        if isinstance(key, bytes):
            return key.decode("utf-8")
        return key

    def get_limiter(self, name: str) -> "Limiter":
        return self.limiters[name]

    def attempt(self, key: str, callback: Callable, max_attempts: int, delay: int = 60):
        # don't execute callback if key limited
        if self.too_many_attempts(key, max_attempts):
            return False

        result = callback()
        self.hit(key, delay)
        return result

    def too_many_attempts(self, key: str, max_attempts: int) -> bool:
        key = self.clean_key(key)
        if self.attempts(key) >= max_attempts:
            # trigger remove of cache value if needed
            self.cache.get(f"{key}:timer")
            if self.cache.has(f"{key}:timer"):
                return True
            self.reset_attempts(key)
        return False

    def hit(self, key: str, delay: int) -> int:
        key = self.clean_key(key)
        # store timestamp when key limit be available again
        available_at = pendulum.now().add(seconds=delay).int_timestamp
        self.cache.add(f"{key}:timer", available_at, delay)
        # ensure key exists
        self.cache.add(key, 0, delay)
        hits = self.cache.increment(key)
        return hits

    def reset_attempts(self, key: str) -> bool:
        key = self.clean_key(key)
        return self.cache.put(key, 0)

    def clear(self, key: str):
        key = self.clean_key(key)
        self.cache.forget(key)
        self.cache.forget(f"{key}:timer")

    def available_at(self, key: str) -> int:
        """Get UNIX integer timestamp at which key will be available again."""
        key = self.clean_key(key)
        timestamp = int(self.cache.get(f"{key}:timer", 0))
        return timestamp

    def available_in(self, key: str) -> int:
        """Get seconds in which key will be available again."""
        timestamp = self.available_at(key)
        if not timestamp:
            return 0
        else:
            return max(0, timestamp - pendulum.now().int_timestamp)

    def remaining(self, key: str, max_attempts: int) -> int:
        """Get remaining attempts before limitation."""
        key = self.clean_key(key)
        return max_attempts - self.attempts(key)



import json
from typing import TYPE_CHECKING, Any


from ..BaseDriver import BaseDriver
from ...validation import MessageBag

if TYPE_CHECKING:
    from ...foundation import Application


class RedisDriver(BaseDriver):
    """Session driver used to store data in Redis."""

    def __init__(self, application: "Application"):
        super().__init__(application)
        self.connection = None

    def get_connection(self):
        try:
            import redis
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'redis' library. Run 'pip install redis' to fix this."
            )

        if not self.connection:
            self.connection = redis.Redis(
                **self.options.get("options", {}),
                host=self.options.get("host"),
                port=self.options.get("port", 6379),
                password=self.options.get("password", None),
                decode_responses=True
            )

        return self.connection

    def start(self) -> dict:
        data = {}
        flashed = {}

        session_data = {}
        cursor = "0"
        session_prefix = self.get_session_namespace() + "*"
        while cursor != 0:
            cursor, keys = self.get_connection().scan(
                cursor=cursor, match=session_prefix, count=100000
            )
            if keys:
                values = self.get_connection().mget(*keys)
                # values = map(int, values)
                session_data.update(dict(zip(keys, values)))

        data_prefix = self.get_data_prefix()
        flash_prefix = self.get_flash_prefix()
        for key, value in session_data.items():
            if key.startswith(data_prefix):
                data.update({key.replace(data_prefix, ""): value})
            elif key.startswith(flash_prefix):
                flashed.update({key.replace(flash_prefix, ""): value})

        return {"data": data, "flashed": flashed}

    def save(
        self,
        added: dict = None,
        deleted: list = None,
        flashed: dict = None,
        deleted_flashed: list = None,
    ) -> None:
        if added is None:
            added = {}
        if deleted is None:
            deleted = []
        if flashed is None:
            flashed = {}
        if deleted_flashed is None:
            deleted_flashed = []

        for key, value in flashed.items():
            if isinstance(value, (MessageBag)):
                value = value.json()
            self.set(f"{self.get_flash_prefix()}{key}", value)

        for key, value in added.items():
            self.set(f"{self.get_data_prefix()}{key}", value)

        for key in deleted:
            self.delete(f"{self.get_data_prefix()}{key}")

        for key in deleted_flashed:
            self.delete(f"{self.get_flash_prefix()}{key}")

    def get_session_namespace(self, session_id: str = None) -> str:
        if not session_id:
            request = self.application.make("request")
            session_id = request.cookie("SESSID")
        namespace = self.options.get("namespace", "")
        namespace += ":" if namespace else ""
        return f"{namespace}session:{session_id}:"

    def get_data_prefix(self) -> str:
        return f"{self.get_session_namespace()}data:"

    def get_flash_prefix(self) -> str:
        return f"{self.get_session_namespace()}flash:"

    def helper(self) -> "RedisDriver":
        """Use to create builtin helper function."""
        return self

    def get(self, key: str, default: str = None) -> str:
        if not self.has(key):
            return default
        return self.get_value(self.get_connection().get(key))

    def set(self, key: str, value: Any, time: int = None):
        time = time or self.get_timeout()

        if isinstance(value, (dict, list)):
            value = json.dumps(value)

        return self.get_connection().set(key, value, ex=time)

    def has(self, key: str) -> bool:
        return self.get_connection().get(key)

    def delete(self, key: str):
        return self.get_connection().delete(key)

    def flush(self, session_id: str = None) -> None:
        """
        Clears all data for the current (or provided) session id
        """

        cursor = "0"
        session_prefix = self.get_session_namespace(session_id) + "*"
        connection = self.get_connection()
        while cursor != 0:
            cursor, keys = connection.scan(
                cursor=cursor, match=session_prefix, count=100000
            )
            if keys:
                connection.delete(*keys)

    def get_timeout(self) -> int:
        # default timeout of session vars is 24 hrs
        return int(self.options.get("timeout", 60 * 60 * 24))

    def get_value(self, value: Any) -> str:
        value = str(value)
        if value.isdigit():
            return str(value)
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value



import os
import uuid

from ..FileStream import FileStream
from ..File import File
from ...utils.filesystem import get_extension


class AmazonS3Driver:
    def __init__(self, application):
        self.application = application
        self.options = {}
        self.connection = None

    def set_options(self, options):
        self.options = options
        return self

    def get_connection(self):
        try:
            import boto3
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'boto3' library. Run 'pip install boto3' to fix this."
            )

        if not self.connection:
            self.connection = boto3.Session(
                aws_access_key_id=self.options.get("client"),
                aws_secret_access_key=self.options.get("secret"),
                region_name=self.options.get("region"),
            )

        return self.connection

    def get_bucket(self):
        return self.options.get("bucket")

    def get_name(self, path, alias):
        extension = get_extension(path)
        return f"{alias}{extension}"

    def put(self, file_path, content):
        self.get_connection().resource("s3").Bucket(self.get_bucket()).put_object(
            Key=file_path, Body=content
        )
        return content

    def put_file(self, file_path, content, name=None):
        file_name = self.get_name(content.name, name or str(uuid.uuid4()))

        if hasattr(content, "get_content"):
            content = content.get_content()

        self.get_connection().resource("s3").Bucket(self.get_bucket()).put_object(
            Key=os.path.join(file_path, file_name), Body=content
        )
        return os.path.join(file_path, file_name)

    def get(self, file_path):
        try:
            return (
                self.get_connection()
                .resource("s3")
                .Bucket(self.get_bucket())
                .Object(file_path)
                .get()
                .get("Body")
                .read()
                .decode("utf-8")
            )
        except self.missing_file_exceptions():
            pass

    def missing_file_exceptions(self):
        import boto3

        return (boto3.exceptions.botocore.errorfactory.ClientError,)

    def exists(self, file_path):
        try:
            self.get_connection().resource("s3").Bucket(self.get_bucket()).Object(
                file_path
            ).get().get("Body").read()
            return True
        except self.missing_file_exceptions():
            return False

    def missing(self, file_path):
        return not self.exists(file_path)

    def stream(self, file_path):
        return FileStream(
            self.get_connection()
            .resource("s3")
            .Bucket(self.get_bucket())
            .Object(file_path)
            .get()
            .get("Body")
            .read(),
            file_path,
        )

    def copy(self, from_file_path, to_file_path):
        copy_source = {"Bucket": self.get_bucket(), "Key": from_file_path}
        self.get_connection().resource("s3").meta.client.copy(
            copy_source, self.get_bucket(), to_file_path
        )

    def move(self, from_file_path, to_file_path):
        self.copy(from_file_path, to_file_path)
        self.delete(from_file_path)

    def prepend(self, file_path, content):
        value = self.get(file_path)
        content = content + value
        self.put(file_path, content)
        return content

    def append(self, file_path, content):
        value = self.get(file_path) or ""
        value += content
        self.put(file_path, content)

    def delete(self, file_path):
        return (
            self.get_connection()
            .resource("s3")
            .Object(self.get_bucket(), file_path)
            .delete()
        )

    def store(self, file, name=None):
        full_path = name or file.hash_path_name()
        self.get_connection().resource("s3").Bucket(self.get_bucket()).put_object(
            Key=full_path, Body=file.stream()
        )
        return full_path

    def make_file_path_if_not_exists(self, file_path):
        if not os.path.isfile(file_path):
            if not os.path.exists(os.path.dirname(file_path)):
                # Create the path to the model if it does not exist
                os.makedirs(os.path.dirname(file_path))

            return True

        return False

    def get_files(self, directory=None):
        bucket = self.get_connection().resource("s3").Bucket(self.get_bucket())

        if directory:
            objects = bucket.objects.all().filter(Prefix=directory)
        else:
            objects = bucket.objects.all()

        files = []
        for my_bucket_object in objects.all():
            if "/" not in my_bucket_object.key:
                files.append(File(my_bucket_object, my_bucket_object.key))

        return files



import json

from ..views import View
from ..controllers import Controller
from ..utils.structures import data_get


class HttpTestResponse:
    def __init__(self, testcase, application, request, response, route):
        self.testcase = testcase
        self.application = application
        self.request = request
        self.response = response
        self.route = route
        self.content = None
        self.status = None
        self.get_response()

    def get_response(self):
        self.content = self.response.get_response_content()
        return self

    def get_content(self):
        """Take care of decoding content if bytes and returns str."""
        return (
            self.content.decode("utf-8")
            if isinstance(self.content, bytes)
            else str(self.content)
        )

    def assertContains(self, content):
        assert (
            content in self.get_content()
        ), f"{content} not found in {self.get_content()}"
        return self

    def assertNotContains(self, content):
        assert content not in self.get_content()
        return self

    def assertContainsInOrder(self, *content):
        response_content = self.get_content()
        index = 0
        for content_string in content:
            found_at_index = response_content.find(content_string, index)
            assert found_at_index != -1
            index = found_at_index + len(content_string)
        return self

    def assertIsNamed(self, name):
        assert (
            self.route.get_name() == name
        ), f"Route name is {self.route.get_name()}. Asserted {name}"
        return self

    def assertIsNotNamed(self, name=None):
        if name is None:
            assert self.route.name is None, "Route has a name: {}".format(
                self.route.name
            )
        else:
            assert (
                self.route.get_name() != name
            ), f"Route name {self.route.get_name()} matches expected {name}"
        return self

    def assertIsStatus(self, status):
        assert self.response.is_status(
            status
        ), f"Status is {self.response.get_status_code()}. Asserted {status}"
        return self

    def assertNotFound(self):
        return self.assertIsStatus(404)

    def assertOk(self):
        return self.assertIsStatus(200)

    def assertCreated(self):
        return self.assertIsStatus(201)

    def assertLimited(self) -> "HttpTestResponse":
        return self.assertIsStatus(429)

    def assertSuccessful(self):
        assert 200 <= self.response.get_status_code() < 300
        return self

    def assertNoContent(self, status=204):
        assert not self.get_content()
        return self.assertIsStatus(status)

    def assertUnauthorized(self):
        return self.assertIsStatus(401)

    def assertForbidden(self):
        return self.assertIsStatus(403)

    def assertError(self):
        return self.assertIsStatus(500)

    def assertHasHeader(self, name, value=None):
        header_value = self.response.header(name)
        assert header_value, f"Could not find the header {name}"
        if value:
            assert (
                value == header_value
            ), f"Header '{name}' does not equal {value} but {header_value}"
        return self

    def assertHeaderMissing(self, name):
        assert not self.response.header(name)
        return self

    def dumpRequestHeaders(self):
        """Dump request headers."""
        self.testcase.dump(self.request.header_bag.to_dict(), "Request Headers")
        return self

    def dumpResponseHeaders(self):
        """Dump response headers."""
        self.testcase.dump(self.response.header_bag.to_dict(), "Response Headers")
        return self

    def ddHeaders(self):
        """Dump request and response headers and die."""
        self.dumpRequestHeaders()
        self.dumpResponseHeaders()
        self.testcase.stop()

    def dumpRequestHeaders(self):
        """Dump request headers."""
        self.testcase.dump(self.request.header_bag.to_dict(), "Request Headers")
        return self

    def dumpResponseHeaders(self):
        """Dump response headers."""
        self.testcase.dump(self.response.header_bag.to_dict(), "Response Headers")
        return self

    def ddHeaders(self):
        """Dump request and response headers and die."""
        self.dumpRequestHeaders()
        self.dumpResponseHeaders()
        self.testcase.stop()

    def assertLocation(self, location):
        return self.assertHasHeader("Location", location)

    def assertRedirect(self, url=None, name=None, params={}):
        # we could assert 301 or 302 code => what if user uses another status code in redirect()
        # here we are sure
        assert self.get_content() == "Redirecting ..."
        if url:
            self.assertLocation(url)
        elif name:
            url = self.response._get_url_from_route_name(name, params)
            self.assertLocation(url)
        return self

    def assertCookie(self, name, value=None):
        assert self.request.cookie_jar.exists(name)
        if value is not None:
            assert self.request.cookie_jar.get(name).value == value
        return self

    def assertPlainCookie(self, name):
        assert self.request.cookie_jar.exists(name)
        assert not self.request.cookie_jar.get(name).secure
        return self

    def assertCookieExpired(self, name):
        self.assertCookie(name)
        assert self.request.cookie_jar.is_expired(name)
        return self

    def assertCookieNotExpired(self, name):
        return not self.assertCookieExpired(name)

    def assertCookieMissing(self, name):
        assert not self.request.cookie_jar.exists(name)
        return self

    def assertSessionHas(self, key, value=None):
        """Assert that session contains the given key with the corresponding value if given.
        The session driver can be specified if necessary."""
        session = self.request.session
        assert session.has(key)
        if value is not None:
            real_value = session.get(key)
            assert (
                real_value == value
            ), f"Value for {key} is {real_value}, expected {value}"
        return self

    def assertSessionMissing(self, key):
        """Assert that session does not contain the given key. The session driver can be specified
        if necessary."""
        assert not self.request.session.get(key)
        return self

    def assertSessionHasErrors(self, keys=[]):
        """Assert that session contains errors for the given list of keys (meaning that each given key
        exists in 'errors' dict in session.) If no keys are given this will assert that the
        sessions has errors without checking specific keys."""
        session = self.request.session
        assert session.has("errors")
        if keys:
            errors = session.get("errors")
            for key in keys:
                assert errors.get(key)
        return self

    def assertSessionHasNoErrors(self, keys=[]):
        """Assert that session does not have any errors (meaning that session does not contain an
        'errors' key or 'errors' key is empty. If a list of keys is given, this will check
        that there are no errors for each of those keys."""
        session = self.request.session
        if not keys:
            assert not session.has("errors")
        else:
            errors = session.get("errors")
            for key in keys:
                assert not errors.get(key)
        return self

    def dumpSession(self):
        """Dump session data."""
        self.testcase.dump(self.application.make("session").all(), "Session Data")
        return self

    def ddSession(self):
        """Dump session data and die."""
        self.dumpSession()
        self.testcase.stop()

    def _ensure_response_has_view(self):
        """Ensure that the response has a view as its original content."""
        if not (self.response.original and isinstance(self.response.original, View)):
            raise ValueError("The response is not a view")

    def assertViewIs(self, name):
        """Assert that request renders the given view name."""
        self._ensure_response_has_view()
        assert (
            self.response.original.template == name
        ), f"Template {self.response.original.template} is not equal to {name}"
        return self

    def assertViewHas(self, key, value=None):
        """Assert that view context contains a given data key (and eventually associated value)."""
        self._ensure_response_has_view()
        value_at_key = data_get(self.response.original.dictionary, key)
        assert value_at_key
        if value:
            assert value_at_key == value
        return self

    def assertViewHasExact(self, keys):
        """Assert that view context contains exactly the data keys (or the complete data dict)."""
        self._ensure_response_has_view()
        if isinstance(keys, list):
            assert set(keys) == set(self.response.original.dictionary.keys()) - set(
                self.response.original._shared.keys()
            )
        else:
            view_data = self.response.original.dictionary
            for key in self.response.original._shared:
                del view_data[key]
            assert keys == view_data
        return self

    def assertViewMissing(self, key):
        """Assert that given data key is not in the view context."""
        self._ensure_response_has_view()
        assert not data_get(self.response.original.dictionary, key)
        return self

    def assertGuest(self, guard="web"):
        assert not self.application.make("auth").guard(guard).user()
        return self

    def assertAuthenticated(self, user=None, guard="test"):
        """Assert that user is authenticated. If a user a given assert that the given is
        authenticated."""
        logged_user = self.application.make("auth").guard(guard).user()
        assert logged_user
        if user:
            assert user.get_id() == logged_user.get_id()
        return self

    def assertHasHttpMiddleware(self, middleware):
        """Assert that the request/response cycle has the given middleware. The HTTP middleware
        class should be given."""
        assert middleware in self.application.make("middleware").http_middleware
        return self

    def assertHasRouteMiddleware(self, middleware):
        """Assert that the route has the given middleware. The registration key of the middleware
        should be used."""
        assert middleware in self.application.make("middleware").route_middleware
        return self

    def assertHasController(self, controller):
        """Assert that route used the given controller. The controller can be a class or
        a string. If it's a string it should be formatted as follow: ControllerName@method"""
        if isinstance(controller, str) and "@" in controller:
            assert self.route.controller == controller
        elif issubclass(controller, Controller):
            assert self.route.controller_class == controller
        else:
            raise ValueError(
                "controller must be a string like YourController@index or a Controller class"
            )
        return self

    def assertRouteHasParameter(self, key, value=None):
        assert key in self.route.url_list, "Route does not contain parameter {key}."
        if value is not None:
            assert self.request.param(key) == str(value)
            pass
        return self

    def _ensure_response_is_json(self):
        """Parse response back from JSON into a dict, if an error happens the response was not
        a JSON string."""
        try:
            return json.loads(self.response.content)
        except ValueError:
            raise ValueError("The response was not JSON serializable")

    def assertJson(self, data={}):
        """Assert that response is JSON and contains the given data dictionary. The assertion will
        pass even if it is not an exact match."""
        response_data = self._ensure_response_is_json()
        assert data.items() <= response_data.items()
        return self

    def assertJsonPath(self, path, value=None):
        """Assert that response is JSON and contains the given path, with eventually the given
        value if provided. The path is a dotted path."""
        response_data = self._ensure_response_is_json()
        data_at_path = data_get(response_data, path)

        assert data_at_path == value, f"'{data_at_path}' does not equal {value}"
        return self

    def assertJsonExact(self, data):
        """Assert that response is JSON and is exactly the given data."""
        response_data = self._ensure_response_is_json()
        assert response_data == data, f"'{response_data}' does not equal {data}"
        return self

    def assertJsonCount(self, count, key=None):
        """Assert that JSON response is JSON and has the given count of keys at root level
        or at the given key."""
        response_data = self._ensure_response_is_json()
        if key is not None:
            response_data = response_data.get(key, {})

        response_count = len(response_data.keys())
        assert (
            response_count == count
        ), f"JSON response count is {response_count}. Asserted {count}."
        return self

    def assertJsonMissing(self, path):
        """Assert that JSON response is JSON and does not contain given path.
        The path can be a dotted path"""
        response_data = self._ensure_response_is_json()
        assert not data_get(
            response_data, path
        ), f"'{response_data}' is not missing from {data_get(response_data, path)}"
        return self



class Pipeline:
    def __init__(self, payload, *args):
        self.payload = payload
        self.args = args

    def through(self, pipe_list, handler="handle"):
        passthrough = self.payload
        for pipe in pipe_list:
            response = getattr(pipe(), handler)(self.payload, *self.args)
            if response != passthrough:
                return False

        return True



from exceptionite import Tab


class DumpsTab(Tab):
    id = "dumps"
    name = "Dumps"
    component = "DumpsTab"
    icon = "CodeIcon"
    advertise_content = True
    empty_msg = "Nothing dumped !"

    def build(self):
        dumps = self.handler.app.make("dumper").get_serialized_dumps()
        return {
            "dumps": dumps,
        }

    def has_content(self):
        return len(self.handler.app.make("dumper").get_dumps()) > 0



import inspect

from ..exceptions import DumpException
from .Dump import Dump


class Dumper:
    def __init__(self, application):
        self.app = application
        self.dumps = []

    def clear(self):
        """Clear all dumped data"""
        self.dumps = []
        return self

    def dd(self, *objects):
        """Dump all provided args and die, raising a DumpException."""
        self._dump(*objects)
        raise DumpException()

    def dump(self, *objects):
        """Dump all provided args and continue code execution. This does not raise a DumpException."""
        dumps = self._dump(*objects)
        # output dumps in console
        for dump in dumps:
            print(dump)
        return dumps

    def get_dumps(self, ascending=False):
        """Get all dumps as Dump objects."""
        if ascending:
            return self.dumps
        else:
            new_dumps = self.dumps.copy()
            new_dumps.reverse()
            return new_dumps

    def last(self):
        """Return last added dump."""
        return self.dumps[-1]

    def get_serialized_dumps(self, ascending=False):
        """Get all dumps as dict."""
        return list(
            map(lambda dump: dump.serialize(), self.get_dumps(ascending=ascending))
        )

    def _dump(self, *objects):
        # get origin of dumped objects (go up 2 frames)
        function, filename, line = (
            inspect.stack()[2].function,
            inspect.stack()[2].filename,
            inspect.stack()[2].lineno,
        )

        # get name of dumped objects (go up 2 frames)
        some = inspect.currentframe().f_back.f_back.f_locals
        names = {}
        for name, var in some.items():
            names.update({str(var): name})
        named_objects = {}
        for obj in objects:
            # for variables dumped without name, use their type
            default = (
                f"<class '{obj.__name__}'>" if inspect.isclass(obj) else str(type(obj))
            )
            name = names.get(str(obj), default)
            named_objects.update({name: obj})

        dump = Dump(
            named_objects,
            function,
            filename,
            line,
        )
        self.dumps.append(dump)
        return self.dumps



from masonite.middleware import Middleware


class __class__(Middleware):
    def before(self, request, response):
        return request

    def after(self, request, response):
        return request



import re
from urllib import parse

from ..utils.str import modularize, removeprefix
from ..exceptions import InvalidRouteCompileException
from ..facades import Loader
from ..controllers import Controller
from ..exceptions import LoaderNotFound


class HTTPRoute:

    _base_casts_map = {
        "int": int,
        "integer": int,
        "string": str,
        "str": str,
        "float": float,
    }

    def __init__(
        self,
        url,
        controller=None,
        request_method=["get"],
        name=None,
        compilers=None,
        controllers_locations=["app.http.controllers"],
        controller_bindings=[],
        **options,
    ):
        if not url.startswith("/"):
            url = "/" + url
        if url.endswith("/") and url != "/":
            url = url[:-1]

        self.url = url
        self.controllers_locations = controllers_locations
        self.controller = controller
        self.controller_class = None
        self.controller_instance = None
        self.controller_method = None
        self._domain = None
        self._name = name
        self._casts_map = {}
        self._active_regex = {}
        self._should_cast = False

        self.request_method = [x.lower() for x in request_method]
        # add HEAD method if GET
        if "get" in self.request_method and "head" not in self.request_method:
            self.request_method.append("head")

        self.list_middleware = []
        self.excluded_middlewares = []
        self.e = None
        self.compilers = compilers or {}
        self._find_controller(controller)
        self.controller_bindings = controller_bindings
        self.compile_route_to_regex()

    def __str__(self):
        return f"<HttpRoute [{self._name}]: {self.url}>"

    def match(self, path, request_method, subdomain=None):

        route_math = (
            re.match(self._compiled_regex, path)
            or re.match(self._compiled_regex_end, path)
        ) and request_method.lower() in self.request_method

        domain_match = subdomain == self._domain

        return route_math and domain_match

    def get_name(self):
        return self._name

    def matches(self, path):
        return re.match(self._compiled_regex, path) or re.match(
            self._compiled_regex_end, path
        )

    def match_name(self, name):
        return name == self._name

    def name(self, name):
        self._name = name
        return self

    def domain(self, subdomain):
        self._domain = subdomain
        return self

    def to_url(self, parameters: dict = {}, query_params: dict = {}) -> str:
        """Transform route to a URL string with the given url and query parameters."""

        # Split the url into a list
        split_url = self.url.split("/")

        # Start beginning of the new compiled url
        compiled_url = "/"

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if "@" in url:
                    url = url.replace("@", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters[url]) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                elif "?" in url:
                    url = url.replace("?", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters.get(url, "/")) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                else:
                    compiled_url += url + "/"

        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith("/"):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if "//" in compiled_url:
            compiled_url = compiled_url.replace("//", "/")

        # Add eventual query parameters
        if query_params:
            compiled_url += "?" + parse.urlencode(query_params)
        return compiled_url

    def _find_controller(self, controller):
        """Find the controller to attach to the route. Look for controller (str or class) in all
        specified controllers_location.

        Arguments:
            controller {string|object} -- String or object controller to search for.

        Returns:
            None
        """
        if controller is None:
            return None
        # If the output specified is a string controller e.g. "WelcomeController@show"
        elif isinstance(controller, str):
            if "@" in controller:
                controller_path, controller_method_str = controller.split("@")
            else:
                controller_path = controller
                controller_method_str = "__call__"

            controller_path = modularize(controller_path).split(".")
            if len(controller_path) > 1:
                controller_name = controller_path.pop()
                prefix_path = ".".join(controller_path)
            else:
                controller_name = controller_path[0]
                prefix_path = ""
            # build a list of all locations where the controller can be found
            # if the controller is defined such as auth.WelcomeController, append the prefix path to
            # the locations
            locations = list(
                map(
                    lambda loc: f"{loc}.{removeprefix(prefix_path, loc)}"
                    if prefix_path
                    else loc,
                    self.controllers_locations,
                )
            )
            try:
                self.controller_class = Loader.find(
                    Controller, locations, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")
        # controller is an instance with a bound method
        elif hasattr(controller, "__self__"):
            _, controller_method_str = controller.__qualname__.split(".")
            self.controller_instance = controller.__self__

        # it's a class or class.method, we don't have to find it, just get the class
        elif hasattr(controller, "__qualname__"):
            if "." in controller.__qualname__:
                controller_name, controller_method_str = controller.__qualname__.split(
                    "."
                )
            else:
                controller_name = controller.__qualname__
                controller_method_str = "__call__"

            try:
                self.controller_class = Loader.get_object(
                    controller.__module__, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")
        # it's a controller instance
        else:
            self.controller_instance = controller
            controller_method_str = "__call__"

        # Set the controller method on class. This is a string
        self.controller_method = controller_method_str

    def get_response(self, app=None):
        from ..response import Response

        # Resolve Controller Constructor
        if self.e:
            print(
                "\033[93mCannot find controller {}. Did you create this one?".format(
                    self.controller
                ),
                "\033[0m",
            )
            raise SyntaxError(str(self.e))

        if app:
            if self.controller_instance:
                controller = self.controller_instance
            else:
                controller = app.resolve(
                    self.controller_class, *self.controller_bindings
                )
            # resolve route parameters
            params = self.extract_parameters(app.make("request").get_path()).values()
            # Resolve Controller Method
            response = app.resolve(getattr(controller, self.controller_method), *params)

            # if request method is HEAD, return a response without content but with
            # headers that would be returned if request's URL was requested with GET method instead
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/HEAD
            if app.make("request").get_request_method().upper() == "HEAD":
                real_response = app.make("response").view(response)
                response = Response(app).status(204)
                response.header_bag = real_response.header_bag
                app.bind("response", response)
            return response

        if self.controller_instance:
            controller = self.controller_instance
        else:
            controller = self.controller_class(*self.controller_bindings)

        return getattr(controller, self.controller_method)()

    def middleware(self, *args):
        """Load a list of middleware to run.

        Returns:
            self
        """
        for arg in args:
            if arg and arg not in self.list_middleware:
                self.list_middleware.append(arg)
        return self

    def prepend_middleware(self, *args):
        """Load a list of middleware to run.

        Returns:
            self
        """
        for arg in args:
            if arg and arg not in self.list_middleware:
                self.list_middleware.insert(0, arg)
        return self

    def get_middlewares(self):
        """Get all the middlewares to run for this route."""
        middlewares = self.list_middleware

        for middleware in self.excluded_middlewares:
            if middleware in middlewares:
                middlewares.remove(middleware)

        return middlewares

    def set_middleware(self, middleware):
        """Get all the middlewares to run for this route."""
        self.list_middleware = middleware

        return self

    def exclude_middleware(self, *args):
        """Remove a list of middleware for this route. It can be useful when
        using Route group middleware to override middleware for a given route in the group."""
        for middleware in args:
            self.excluded_middlewares.append(middleware)
        return self

    def compile_route_to_regex(self):
        """Compile the given route to a regex string.

        Arguments:
            route {string} -- URI of the route to compile.

        Returns:
            string -- Compiled URI string.
        """
        # Split the route
        split_given_route = self.url.split("/")
        # compile the provided url into regex
        url_list = []
        regex = "^"
        for regex_route in split_given_route:
            if "@" in regex_route:
                if ":" in regex_route:
                    try:
                        param_name, compiler_name = regex_route.split(":")
                        regex += self.compilers[compiler_name]
                        self._active_regex.update(
                            {param_name.replace("@", ""): compiler_name}
                        )
                    except KeyError:
                        raise InvalidRouteCompileException(
                            'Route compiler "{}" is not an available route compiler. '
                            "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                regex_route.split(":")[1]
                            )
                        )

                else:
                    regex += self.compilers["default"]

                regex += r"\/"

                # append the variable name passed @(variable):int to a list
                url_list.append(regex_route.replace("@", "").split(":")[0])
            elif "?" in regex_route:
                # Make the preceding token match 0 or more
                regex += "?"

                if ":" in regex_route:

                    try:
                        param_name, compiler_name = regex_route.split(":")
                        regex += self.compilers[compiler_name] + "*"
                        self._active_regex.update(
                            {param_name.replace("@", ""): compiler_name}
                        )
                    except KeyError:
                        if self.request:
                            raise InvalidRouteCompileException(
                                'Route compiler "{}" is not an available route compiler. '
                                "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                    regex_route.split(":")[1]
                                )
                            )
                        self._compiled_regex = None
                        self._compiled_regex_end = None
                        return

                else:
                    regex += self.compilers["default"] + "*"

                regex += r"\/"

                url_list.append(regex_route.replace("?", "").split(":")[0])
            else:
                regex += regex_route + r"\/"

        self.url_list = url_list
        regex += "$"
        self._compiled_regex = re.compile(regex.replace(r"\/$", r"$"))
        self._compiled_regex_end = re.compile(regex)

        return regex

    def extract_parameters(self, path):
        if not self.url_list:
            return {}

        if (not path.endswith("/")) or path == "/":
            matching_regex = self._compiled_regex
        else:
            matching_regex = self._compiled_regex_end
        params = dict(zip(self.url_list, matching_regex.match(path).groups()))

        if self._should_cast:
            for param, value in params.items():
                # get cast type from cast map if defined else use cast type from compiler else str
                cast_type = self._casts_map.get(
                    param,
                    self._base_casts_map.get(self._active_regex.get(param), str),
                )
                params.update({param: cast_type(value)})
        return params

    def casts(self, _casts_map: dict = {}) -> "HTTPRoute":
        self._casts_map = _casts_map
        self._should_cast = True
        return self



from .ExceptionHandler import ExceptionHandler
from .handlers.DumpExceptionHandler import DumpExceptionHandler
from .handlers.HttpExceptionHandler import HttpExceptionHandler
from .handlers.ModelNotFoundHandler import ModelNotFoundHandler
from .DD import DD
from .exceptions import (
    AuthorizationException,
    InvalidRouteCompileException,
    RouteMiddlewareNotFound,
    ContainerError,
    MissingContainerBindingNotFound,
    StrictContainerException,
    ResponseError,
    InvalidHTTPStatusCode,
    RequiredContainerBindingNotFound,
    ViewException,
    RouteNotFoundException,
    DumpException,
    InvalidSecretKey,
    InvalidCSRFToken,
    NotificationException,
    InvalidToken,
    ProjectLimitReached,
    ProjectProviderTimeout,
    ProjectProviderHttpError,
    ProjectTargetNotEmpty,
    MixFileNotFound,
    MixManifestNotFound,
    InvalidConfigurationLocation,
    InvalidConfigurationSetup,
    InvalidPackageName,
    LoaderNotFound,
    QueueException,
    AmbiguousError,
    MethodNotAllowedException,
    ModelNotFoundException,
    ThrottleRequestsException,
)



import os

from .. import Middleware
from ...utils.location import base_path
from ...facades import View


class MaintenanceModeMiddleware(Middleware):
    def before(self, request, response):
        down = os.path.exists(base_path(".down"))
        if down is True:
            return response.view(View.render("maintenance"), status=503)
        return request

    def after(self, request, _):
        return request



from .Mail import Mail
from ..configuration import config


class MockMail(Mail):
    def __init__(self, application, *args, **kwargs):
        super().__init__(application, *args, **kwargs)
        self.count = 0
        self.driver = None

    def reset(self):
        """Reset mock implementation."""
        self.count = 0
        self.driver = None

    def send(self, driver=None):
        if driver:
            self.driver = driver
        else:
            self.driver = self.options.get("driver", None) or config(
                "mail.drivers.default"
            )

        config_options = self.get_config_options(self.driver)
        if self.options.get("from"):
            config_options.pop("from", None)
        self.options.update(config_options)
        self.count += 1
        return self

    def seeEmailBcc(self, bcc):
        assert bcc == self.options.get(
            "bcc"
        ), f"BCC of {self.options.get('bcc')} does not match expected {bcc}"
        return self

    def seeEmailCc(self, cc):
        assert cc == self.options.get(
            "cc"
        ), f"CC of {self.options.get('cc')} does not match expected {cc}"
        return self

    def seeEmailContains(self, contents):
        assert contents in self.options.get(
            "html_content"
        ) or contents in self.options.get(
            "text_content"
        ), f"Could not find the {contents} in the email"
        return self

    def getHtmlContents(self, contents):
        return self.options.get("html_content")

    def getTextContents(self, contents):
        return self.options.get("text_content")

    def seeEmailCountEquals(self, count):
        assert (
            count == self.count
        ), f"Email count of {self.count} does not match expected {count}"
        return self

    def seeEmailDoesNotContain(self, contents):
        assert contents not in self.options.get(
            "html_content"
        ) and contents not in self.options.get(
            "text_content"
        ), f"Found {contents} in the email but should not be"
        return self

    def seeEmailFrom(self, assertion):
        assert assertion == self.options.get(
            "from"
        ), f"Assertion of from address {self.options.get('from')} does not match expected {assertion}"
        return self

    def seeEmailReplyTo(self, assertion):
        assert assertion == self.options.get(
            "reply_to"
        ), f"Assertion of reply-to {self.options.get('reply_to')} does not match expected {assertion}"
        return self

    def seeEmailSubjectContains(self, assertion):
        assert assertion in self.options.get(
            "subject"
        ), f"Assertion of subject {self.options.get('subject')} does not contain expected {assertion}"
        return self

    def seeEmailSubjectDoesNotContain(self, assertion):
        assert assertion not in self.options.get(
            "subject"
        ), f"Assertion of subject {self.options.get('subject')} does contain expected {assertion}"
        return self

    def seeEmailSubjectEquals(self, assertion):
        assert assertion == self.options.get(
            "subject"
        ), f"Assertion of subject address {self.options.get('subject')} does not match expected {assertion}"
        return self

    def seeEmailTo(self, assertion):
        assert assertion == self.options.get(
            "to"
        ), f"Assertion of to address {self.options.get('to')} does not match expected {assertion}"
        return self

    def seeEmailPriority(self, assertion):
        assert assertion == self.options.get(
            "priority"
        ), f"Assertion of priority {self.options.get('priority')} does not match expected {assertion}"
        return self

    def seeEmailWasNotSent(self):
        assert self.count == 0, "Expected email was not sent but it was sent"
        return self

    def seeEmailWasSent(self):
        assert self.count > 0, "Expected email was not sent but it was sent"
        return self

    def seeDriverWas(self, name):
        assert (
            self.driver == name
        ), f"Expected email driver to be {name} but it was {self.driver}"
        return self



"""View Module."""
from collections import defaultdict
from os.path import split, exists
from jinja2 import (
    ChoiceLoader,
    Environment,
    PackageLoader,
    select_autoescape,
    BaseLoader,
)
from jinja2.exceptions import TemplateNotFound
from typing import TYPE_CHECKING, Callable, Any
from ..exceptions import ViewException
from ..utils.str import as_filepath
from ..utils.location import views_path


if TYPE_CHECKING:
    from ..foundation import Application


def path_to_package(path, separator="/"):
    # ensure no leading/trailing slashes before splitting to avoid blank strings
    location = path.strip(separator).split(separator)
    package_name = location[0]
    package_path = "/".join(location[1:])
    return package_name, package_path


class View:
    """Responsible for handling everything involved with views and view environments."""

    separator = "/"
    extension = ".html"

    def __init__(self, application: "Application"):
        self.application = application

        # specific to given view rendering
        self.dictionary = {}
        self.composers = {}
        self.template = None
        self.loaders = []
        self.namespaces = defaultdict(list)
        self.env = None
        self._jinja_extensions = ["jinja2.ext.loopcontrols"]
        self._filters = {}
        self._shared = {}
        self._tests = {}

    def render(self, template: str, dictionary: dict = None) -> "View":
        if not dictionary:
            dictionary = {}
        """Render the given template name with the given context as string."""
        if not isinstance(dictionary, dict):
            raise ViewException(
                "Second parameter to render method needs to be a dictionary, {} passed.".format(
                    type(dictionary).__name__
                )
            )

        self.load_template(template)

        # prepare template context
        self.dictionary = {}
        self.dictionary.update(dictionary)
        self.dictionary.update(self._shared)
        if self.composers:
            self.hydrate_from_composers()

        if self._tests:
            self.env.tests.update(self._tests)

        self.rendered_template = self._render()

        return self

    def get_content(self) -> str:
        """Get the rendered content as string."""
        return self.rendered_template

    def _render(self):
        try:
            # Try rendering the template with '.html' appended
            return self.env.get_template(self.filename).render(self.dictionary)
        except TemplateNotFound:
            # Try rendering the direct template the user has supplied
            try:
                return self.env.get_template(self.template).render(self.dictionary)
            except TemplateNotFound as e:
                # Try rendering the direct template the user has supplied
                if self.exists(self.template):
                    raise TemplateNotFound(
                        f"One of the included templates in the '{self.template}' view could not be found"
                    )
                else:
                    raise TemplateNotFound(
                        f"Template '{self.template}' not found"
                    ) from e

    def hydrate_from_composers(self):
        """Add data into the view from specified composers."""
        # Check if the template is directly specified in the composer
        if self.template in self.composers:
            self.dictionary.update(self.composers.get(self.template))

        # Check if there is just an asterisks in the composer
        if "*" in self.composers:
            self.dictionary.update(self.composers.get("*"))

        # We will append onto this string for an easier way to search through wildcard routes
        compiled_string = ""

        # Check for wildcard view composers
        for template in self.template.split(self.separator):
            # Append the template onto the compiled_string
            compiled_string += template
            if self.composers.get("{}*".format(compiled_string)):
                self.dictionary.update(self.composers["{}*".format(compiled_string)])
            else:
                # Add a slash to symbolize going into a deeper directory structure
                compiled_string += "/"

    def composer(self, composer_name: str, dictionary: dict) -> "View":
        """Add/Update composer with the given name and data."""
        if isinstance(composer_name, str):
            self.composers[composer_name] = dictionary

        if isinstance(composer_name, list):
            for composer in composer_name:
                self.composers[composer] = dictionary

        return self

    def share(self, dictionary: dict) -> "View":
        """Share data to all templates."""
        self._shared.update(dictionary)
        return self

    def exists(self, template: str) -> bool:
        """Check if a template with the given name exists."""
        self.load_template(template)

        try:
            self.env.get_template(self.filename)
            return True
        except TemplateNotFound:
            return False

    def add_location(
        self, template_location: str, loader: "BaseLoader" = PackageLoader
    ):
        """Add location directory from which view templates can be loaded. The Jinja2 loader type
        can be specified."""
        if loader == PackageLoader:
            package_name, package_path = path_to_package(template_location)
            self.loaders.append(loader(package_name, package_path))
        else:
            self.loaders.append(loader(template_location))

    def add_namespaced_location(self, namespace: str, template_location: str):
        """Add namespaced location directory from which view templates can be loaded."""
        # if views have been published, add the published view directory as a location
        published_path = views_path(f"vendor/{namespace}/", absolute=False)
        if exists(published_path):
            self.namespaces[namespace].append(
                views_path(f"vendor/{namespace}/", absolute=False)
            )
        # put this one in 2nd as project views must be used first to be able to override package views
        self.namespaces[namespace].append(template_location)

    def add_from_package(self, package_name: str, path_in_package: str):
        self.environments.append(PackageLoader(package_name, path_in_package))

    def filter(self, name: str, function: Callable):
        """Add filter functions to views with the given name."""
        self._filters.update({name: function})

    def add_extension(self, extension: str) -> "View":
        """Register Jinja2 extension to views."""
        self._jinja_extensions.append(extension)
        return self

    def load_template(self, template: str):
        """Private method for loading all the locations into the current environment."""
        self.template = template
        # transform given template path into a real file path with the configured extension
        self.filename = (
            as_filepath(template).replace(self.extension, "") + self.extension
        )
        # assess if new loaders are required for the given template
        template_loaders = []
        # Case 1: the templates needs to be loaded from a namespace
        if ":" in template:
            namespace, rel_template_path = template.split(":")
            self.filename = (
                as_filepath(rel_template_path).replace(self.extension, "")
                + self.extension
            )
            namespace_paths = self.namespaces.get(namespace, None)
            if not namespace_paths:
                raise Exception(f"No such view namespace {namespace}.")
            for namespace_path in namespace_paths:
                package_name, package_path = path_to_package(namespace_path)
                template_loaders.append(PackageLoader(package_name, package_path))

        # Case 2: an absolute path has been given
        elif template.startswith("/"):
            directory, filename = split(template)
            self.filename = filename.replace(self.extension, "") + self.extension
            package_name, package_path = path_to_package(directory)
            template_loaders.append(PackageLoader(package_name, package_path))

        # Else: use already defined view locations to load this template
        loader = ChoiceLoader(template_loaders + self.loaders)

        # @josephmancuso: what is this ??
        # Set the searchpath since some packages look for this object
        # This is sort of a hack for now
        loader.searchpath = ""

        self.env = Environment(
            loader=loader,
            autoescape=select_autoescape(["html", "xml"]),
            extensions=self._jinja_extensions,
            line_statement_prefix="@",
        )
        # add container to environment so that extensions can use it
        self.env.application = self.application

        # add filters to environment
        self.env.filters.update(self._filters)

    def get_current_loaders(self):
        """Get all enabled Jinja2 loaders."""
        if self.env:
            return self.env.loader.loaders

    def set_separator(self, token: str) -> "View":
        """Change separator for view names (default is /)."""
        self.separator = token
        return self

    def set_file_extension(self, extension: str) -> "View":
        """Change file view extension (default is .html)."""
        self.extension = extension
        return self

    def get_response(self) -> str:
        """Get the rendered content as string."""
        return self.rendered_template

    def test(self, key: str, obj: Any) -> "View":
        self._tests.update({key: obj})
        return self



"""Helpers for multiple data structures"""
import importlib
from importlib.abc import Loader
from dotty_dict import dotty

from ..exceptions.exceptions import LoaderNotFound

from .str import modularize


def load(path, object_name=None, default=None, raise_exception=False):
    """Load the given object from a Python module located at path and returns a default
    value if not found. If no object name is provided, loads the module.

    Arguments:
        path {str} -- A file path or a dotted path of a Python module
        object {str} -- The object name to load in this module (None)
        default {str} -- The default value to return if object not found in module (None)
    Returns:
        {object} -- The value (or default) read in the module or the module if no object name
    """
    # modularize path if needed
    module_path = modularize(path)
    # module = pydoc.locate(dotted_path)
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError:
        if raise_exception:
            raise LoaderNotFound(
                f"{module_path} not found or error when importing this module."
            )
        return None

    if object_name is None:
        return module
    else:
        try:
            return getattr(module, object_name)
        except KeyError:
            if raise_exception:
                raise LoaderNotFound(f"{object_name} not found in {module_path}")
            else:
                return default


def data(dictionary={}):
    """Transform the given dictionary to be read/written with dot notation.

    Arguments:
        dictionary {dict} -- a dictionary structure

    Returns:
        {dict} -- A dot dictionary
    """
    return dotty(dictionary)


def data_get(dictionary, key, default=None):
    """Read dictionary value from key using nested notation.

    Arguments:
        dictionary {dict} -- a dictionary structure
        key {str} -- the dotted (or not) key to look for
        default {object} -- the default value to return if the key does not exist (None)

    Returns:
        value {object}
    """
    # dotty dict uses : instead of * for wildcards
    dotty_key = key.replace("*", ":")
    return data(dictionary).get(dotty_key, default)


def data_set(dictionary, key, value, overwrite=True):
    """Set dictionary value at key using nested notation. Values are overridden by default but
    this behaviour can be changed by passing overwrite=False.
    The dictionary is edited in place but is also returned.

    Arguments:
        dictionary {dict} -- a dictionary structure
        key {str} -- the dotted (or not) key to set
        value {object} -- the value to set at key
        overwrite {bool} -- override the value if key exists in dictionary (True)

    Returns:
        dictionary {dict} -- the edited dictionary
    """
    if "*" in key:
        raise ValueError("You cannot set values with wildcards *")
    if not overwrite and data_get(dictionary, key):
        return
    data(dictionary)[key] = value
    return dictionary



class PublishableResource:
    def __init__(self, key):
        self.key = key
        self.files = []

    def add(self, source, destination):
        self.files.append((source, destination))
        return self

    # def add(self, *resources):
    #     for source, destination in resources:
    #         self.sources.append(source)
    #         self.destinations.append(destination)
    #     return self



import json
from os.path import join, exists

from ..configuration import config
from ..utils.location import base_path
from ..exceptions import MixManifestNotFound, MixFileNotFound


class MixHelper:
    def __init__(self, app):
        self.app = app

    def url(self, path, manifest_dir=""):
        if not path.startswith("/"):
            path = "/" + path

        root_url = config("application.mix_base_url") or config("application.app_url")

        # load manifest file
        manifest_file = base_path(join(manifest_dir, "mix-manifest.json"))
        if not exists(manifest_file):
            raise MixManifestNotFound(
                "Mix manifest file mix-manifest.json does not exist."
            )
        manifest = {}
        with open(manifest_file, "r") as f:
            manifest = json.load(f)

        # build asset path
        try:
            compiled_path = manifest[path]
        except KeyError:
            raise MixFileNotFound(f"Can't locate mix file: {path}.")
        return join(root_url, compiled_path.lstrip("/"))



"""DatabaseNotification Model."""
import pendulum
from masoniteorm.relationships import morph_to
from masoniteorm.models import Model


class DatabaseNotification(Model):
    """DatabaseNotification Model allowing notifications to be stored in database."""

    __fillable__ = ["id", "type", "data", "read_at", "notifiable_id", "notifiable_type"]
    __table__ = "notifications"

    @morph_to("notifiable_type", "notifiable_id")
    def notifiable(self):
        """Get the notifiable entity that the notification belongs to."""
        return

    def mark_as_read(self):
        """Mark the notification as read."""
        if not self.read_at:
            self.read_at = pendulum.now()
            return self.save(query=True)

    def mark_as_unread(self):
        """Mark the notification as unread."""
        if self.read_at:
            self.read_at = None
            return self.save(query=True)

    @property
    def is_read(self):
        """Determine if a notification has been read."""
        return self.read_at is not None

    @property
    def is_unread(self):
        """Determine if a notification has not been read yet."""
        return self.read_at is None



import json
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from redis import Redis


class RedisDriver:
    def __init__(self, application):
        self.application = application
        self.connection = None
        self._internal_cache: "dict|None" = None

    def set_options(self, options: dict) -> "RedisDriver":
        self.options = options
        return self

    def get_connection(self) -> "Redis":
        try:
            from redis import Redis
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'redis' library. Run 'pip install redis' to fix this."
            )

        if not self.connection:
            self.connection = Redis(
                **self.options.get("options", {}),
                host=self.options.get("host"),
                port=self.options.get("port"),
                password=self.options.get("password"),
                decode_responses=True,
            )

        # populate the internal cache the first time
        # the connection is established
        if self._internal_cache is None and self.connection:
            self._load_from_store(self.connection)

        return self.connection

    def _load_from_store(self, connection: "Redis" = None) -> None:
        """
        copy all the "cache" key value pairs for faster access
        """
        if not connection:
            return

        if self._internal_cache is None:
            self._internal_cache = {}

        cursor = "0"
        prefix = self.get_cache_namespace()
        while cursor != 0:
            cursor, keys = connection.scan(
                cursor=cursor, match=prefix + "*", count=100000
            )
            if keys:
                values = connection.mget(*keys)
                store_data = dict(zip(keys, values))
                for key, value in store_data.items():
                    key = key.replace(prefix, "")
                    value = self.unpack_value(value)
                    self._internal_cache.setdefault(key, value)

    def get_cache_namespace(self) -> str:
        """
        Build the prefix for the key to use in Redis
        """
        namespace = self.options.get("name", None)
        namespace += ":" if namespace else ""
        return f"{namespace}cache:"

    def add(self, key: str, value: Any = None) -> Any:
        if not value:
            return None

        self.put(key, value)
        return value

    def get(self, key: str, default: Any = None, **options) -> Any:
        if default and not self.has(key):
            self.put(key, default, **options)
            return default

        return self._internal_cache.get(key)

    def put(self, key: str, value: Any = None, seconds: int = None, **options) -> Any:
        if not key or value is None:
            return None

        time = self.get_expiration_time(seconds)

        store_value = value
        if isinstance(value, (dict, list, tuple)):
            store_value = json.dumps(value)

        self.get_connection().set(
            f"{self.get_cache_namespace()}{key}", store_value, ex=time
        )

        if not self.has(key):
            self._internal_cache.update({key: value})

    def has(self, key: str) -> bool:
        return key in self._internal_cache

    def increment(self, key: str, amount: int = 1) -> int:
        value = int(self.get(key)) + amount
        self.put(key, value)
        return value

    def decrement(self, key: str, amount: int = 1) -> int:
        value = int(self.get(key)) - amount
        self.put(key, value)
        return value

    def remember(self, key: str, callable):
        value = self.get(key)

        if value:
            return value

        callable(self)

    def forget(self, key: str) -> None:
        self.get_connection().delete(f"{self.get_cache_namespace()}{key}")
        self._internal_cache.pop(key)

    def flush(self) -> None:
        return self.get_connection().flushall()

    def get_expiration_time(self, seconds: int) -> int:
        if seconds is None:
            seconds = 31557600 * 10

        return seconds

    def unpack_value(self, value: Any) -> Any:
        value = str(value)
        if value.isdigit():
            return str(value)
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value



import inspect
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed

from ...exceptions import QueueException


class AsyncDriver:
    def __init__(self, application):
        self.application = application

    def set_options(self, options):
        self.options = options
        return self

    def push(self, *jobs, args=(), **kwargs):
        """Push objects onto the async stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
            options {**kwargs of options} - Additional options for async driver
        """

        # Initialize Extra Options
        options = self.options
        callback = options.get("callback", "handle")
        mode = options.get("mode", "threading")
        workers = options.get("workers", None)

        # Set processor to either use threads or processes
        processor = self._get_processor(mode=mode, max_workers=workers)
        is_blocking = options.get("blocking", False)

        ran = {}
        for obj in jobs:
            obj = self.application.resolve(obj) if inspect.isclass(obj) else obj
            try:
                future = processor.submit(getattr(obj, callback), *args, **kwargs)
            except AttributeError:
                # Could be wanting to call only a method asynchronously
                future = processor.submit(obj, *args, **kwargs)
            ran.update({future: obj})

        if is_blocking:
            for job in as_completed(ran.keys()):
                if job.exception():
                    ran[job].failed(ran[job], job.exception())
                print(f"Job Ran: {job}")

    def consume(self, **options):
        pass

    def retry(self, **options):
        pass

    def _get_processor(self, mode, max_workers):
        """Set processor to use either threads or multiprocesses

        Arguments:
            mode {str} - async mode
            max_workers {int} - number of threads/processes to use
        """

        if max_workers is None:
            # Use this number because ThreadPoolExecutor is often
            # used to overlap I/O instead of CPU work.
            max_workers = (os.cpu_count() or 1) * 5

        # Determine Mode for Processing
        if mode == "threading":
            processor = ThreadPoolExecutor(max_workers)
        elif mode == "multiprocess":
            processor = ProcessPoolExecutor(max_workers)
        else:
            raise QueueException("Queue mode {} not recognized".format(mode))
        return processor



class Argon2Hasher:
    def __init__(self, options={}):
        self.options = options

    def set_options(self, options):
        self.options = options
        return self

    def _get_password_hasher(self):
        try:
            import argon2
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'argon2' library. Run 'pip install argon2-cffi' to fix this."
            )

        memory = self.options.get("memory", argon2.DEFAULT_MEMORY_COST)
        threads = self.options.get("threads", argon2.DEFAULT_PARALLELISM)
        time = self.options.get("time", argon2.DEFAULT_TIME_COST)
        return argon2.PasswordHasher(
            memory_cost=memory, parallelism=threads, time_cost=time
        )

    def make(self, string):
        return self.make_bytes(string).decode("utf-8")

    def make_bytes(self, string):
        ph = self._get_password_hasher()
        return bytes(ph.hash(bytes(string, "utf-8")), "utf-8")

    def check(self, plain_string, hashed_string):
        ph = self._get_password_hasher()
        return ph.verify(hashed_string, bytes(plain_string, "utf-8"))

    def needs_rehash(self, hashed_string):
        ph = self._get_password_hasher()
        return ph.check_needs_rehash(hashed_string)



import os

from .Command import Command


class UpCommand(Command):
    """
    Brings the server out of maintenance state.

    up
    """

    def handle(self):
        os.remove(".down")
        self.info("Server is online again !")



class HttpExceptionHandler:
    def __init__(self, application):
        self.application = application

    def handle(self, exception):
        status_code = exception.get_status()
        view_name = f"errors/{status_code}"
        response = self.application.make("response")
        request = self.application.make("request")

        if request.accepts_json():
            payload = {
                "status": exception.get_status(),
                "message": exception.get_response(),
            }
            return response.json(payload, status_code)

        # Renders HTTP exception as HTML with predefined error page if exists
        if self.application.make("view").exists(view_name):
            return response.view(
                self.application.make("view").render(
                    f"errors/{status_code}", {"message": exception.get_response()}
                ),
                status_code,
            )
        else:
            # Else render the exception without using template
            return response.view(exception.get_response(), status_code)



import json
from typing import TYPE_CHECKING, Any

from ..validation import MessageBag
from ..exceptions import InvalidConfigurationSetup

if TYPE_CHECKING:
    from ..foundation import Application


class Session:
    """Session manager which provides a way to store information in a persistent store / backend
    that can be accessed from subsequent requests."""

    def __init__(self, application: "Application", driver_config: dict = None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self._active_driver = None
        self.options = {}
        self.data = {}
        self.added = {}
        self.flashed = {}
        self.deleted = []
        self.deleted_flashed = []

    def add_driver(self, name: str, driver: Any) -> None:
        """Register a new session driver with the given name."""
        driver.set_options(self.get_config_options(name))
        self.drivers.update({name: driver})

    def driver(self, driver: str) -> Any:
        """Get a registered session driver with the given name."""
        return self.drivers[driver]

    def set_configuration(self, config: dict) -> "Session":
        """Set session driver options."""

        # make sure the default driver is defined
        if "default" not in config:
            raise InvalidConfigurationSetup("'default' session driver is not defined.")
        # and has a config
        if config["default"] not in config:
            raise InvalidConfigurationSetup(
                f"'{config['default']}' session driver configuration not defined."
            )

        self.driver_config = config
        return self

    def get_driver(self, name: str = None) -> Any:
        """Get the default session driver or the driver with the given name."""
        if name is None:
            return self.drivers[self._active_driver]

        return self.drivers[name]

    def get_config_options(self, driver: str = None) -> dict:
        """Get the options of the default session driver or of the driver with the given name."""
        if driver is None:
            return self.driver_config.get(self._active_driver)

        return self.driver_config.get(driver, {})

    # Start of methods
    def start(self, driver: str = None) -> "Session":
        """Initialize session."""
        self.added = {}
        self.deleted = []
        self.deleted_flashed = []
        self._active_driver = driver or self.get_config_options("default")
        started_data = self.get_driver(name=self._active_driver).start()
        self.data = started_data.get("data", {})
        self.flashed = started_data.get("flashed", {})
        return self

    def get_data(self) -> dict:
        """Get all session data."""
        data = self.data
        data.update(self.added)
        data.update(self.flashed)
        for deleted in self.deleted:
            if deleted in data:
                data.pop(deleted)
        for deleted in self.deleted_flashed:
            if deleted in data:
                data.pop(deleted)
        return data

    def save(self, driver: str = None) -> None:
        """Save session data for the default session driver or the given named driver."""
        return self.get_driver(name=driver).save(
            added=self.added,
            deleted=self.deleted,
            flashed=self.flashed,
            deleted_flashed=self.deleted_flashed,
        )

    def set(self, key: str, value: Any) -> None:
        """Save value in default session."""
        try:
            if isinstance(value, (dict, list, int)) or (
                isinstance(value, str) and value.isnumeric()
            ):
                value = json.dumps(value)
        except json.decoder.JSONDecodeError:
            pass

        self.added.update({key: value})
        self.save()

    def increment(self, key: str, count: int = 1) -> None:
        """Increment session key with given count."""
        return self.set(key, str(int(self.get(key)) + count))

    def decrement(self, key: str, count: int = 1) -> None:
        """Decrement session key with given count."""
        return self.set(key, str(int(self.get(key)) - count))

    def has(self, key: str) -> bool:
        """Check if key is present in active session."""
        return key in self.added or key in self.flashed or key in self.data

    def get(self, key: str) -> Any:
        """Get value of the given key in active session."""
        if key in self.flashed:
            value = self.flashed.get(key)

            try:
                if value is not None and not isinstance(value, MessageBag):
                    value = json.loads(value)
            except json.decoder.JSONDecodeError:
                pass
            self.flashed.pop(key)
            self.deleted_flashed.append(key)
            self.save()
            return value

        value = self.get_data().get(key)
        try:
            if value is not None and not isinstance(value, MessageBag):
                value = json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
        return value

    def pull(self, key: str) -> Any:
        """Get and remove value for the given key in session."""
        key_value = self.get(key)
        self.delete(key)
        return key_value

    def flush(self) -> None:
        """Delete all keys from session."""
        self.deleted += list(self.get_data().keys())
        self.save()

    def delete(self, key: str) -> "None|Any":
        """Delete the given key from session."""
        self.deleted.append(key)
        if key in self.flashed:
            self.flashed.pop(key)
        self.save()

    def flash(self, key: str, value: Any) -> None:
        """Save temporary value into session."""
        try:
            if isinstance(value, (dict, list, int)) or (
                isinstance(value, str) and value.isnumeric()
            ):
                value = json.dumps(value)
        except json.decoder.JSONDecodeError:
            pass

        self.flashed.update({key: value})
        self.save()

    def all(self) -> dict:
        """Get all session data."""
        return self.get_data()



"""New Test Command."""
import inflection
import os

from ..utils.filesystem import make_directory, render_stub_file, get_module_dir
from .Command import Command


class MakeTestCommand(Command):
    """
    Creates a new test class.

    test
        {name : Name of the test case (CamelCase) }
        {--d|--directory=tests/unit : Directory to create the test file}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        test_filename = f"test_{inflection.underscore(self.argument('name').replace('Test', '').replace('test', ''))}.py"
        directory = self.option("directory")
        content = render_stub_file(self.get_testcase_path(), name)

        filepath = os.path.join(directory, test_filename)
        make_directory(filepath)
        if os.path.exists(filepath) and not self.option("force"):
            self.warning(
                f"{filepath} already exists! Run the command with -f (force) to override."
            )
            return -1
        with open(filepath, "w") as f:
            f.write(content)
        self.info(f"Test Created ({filepath})")

    def get_testcase_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/tests/TestCase.py")



from ..facades import View


class ViewController:
    def __init__(self, template, data):
        self.template = template
        self.data = data

    def show(self):
        return View.render(self.template, self.data)



"""Tailwind Preset"""
import shutil

from ..utils.filesystem import make_full_directory
from ..utils.location import base_path, resources_path
from .Preset import Preset


class Tailwind(Preset):
    """
    Configure the front-end scaffolding for the application to use Tailwind
    """

    key = "tailwind"
    packages = {"tailwindcss": "^3.0.7", "postcss": "^8.4.5", "autoprefixer": "^10.4.0"}

    def install(self):
        """Install the preset"""
        self.update_packages(dev=True)
        self.update_webpack_mix()
        self.add_tailwind_config()
        self.update_css()
        self.remove_node_modules()

    def add_tailwind_config(self):
        """Copy example Tailwind configuration into application."""
        shutil.copyfile(
            self.get_template_path("tailwind.config.js"),
            base_path("tailwind.config.js"),
        )

    def update_css(self):
        """Create/Override an app.css file configured for the preset."""
        make_full_directory(resources_path("css"))
        shutil.copyfile(
            self.get_template_path("app.css"), resources_path("css/app.css")
        )



"""A Validation Service Provider."""
from ...providers import Provider
from .. import Validator, ValidationFactory, MessageBag
from ..commands.MakeRuleEnclosureCommand import MakeRuleEnclosureCommand
from ..commands.MakeRuleCommand import MakeRuleCommand


class ValidationProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        validator = Validator()
        self.application.bind("validator", validator)
        self.application.make("commands").add(
            MakeRuleEnclosureCommand(self.application),
            MakeRuleCommand(self.application),
        )

        validator.extend(ValidationFactory().registry)

    def boot(self):
        pass



"""New Key Command."""
import inflection
import os

from ..utils.filesystem import make_directory, render_stub_file, get_module_dir
from ..utils.location import jobs_path
from .Command import Command


class MakeJobCommand(Command):
    """
    Creates a new job class.

    job
        {name : Name of the job}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        content = render_stub_file(self.get_jobs_path(), name)

        filename = f"{name}.py"
        filepath = jobs_path(filename)
        make_directory(filepath)
        if os.path.exists(filepath) and not self.option("force"):
            self.warning(
                f"{filepath} already exists! Run the command with -f (force) to override."
            )
            return -1
        with open(filepath, "w") as f:
            f.write(content)

        # add class to __init__.py
        with open(os.path.join(os.path.dirname(filepath), "__init__.py"), "a") as f:
            f.write(f"from .{name} import {name}\n")

        self.info(f"Job Created ({jobs_path(filename, absolute=False)})")

    def get_template_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/templates/")

    def get_jobs_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/jobs/Job.py")



from os.path import join
from typing import TYPE_CHECKING

from ..utils.str import add_query_params
from ..configuration import config

if TYPE_CHECKING:
    from ..foundation import Application


class UrlsHelper:
    """URLs helper provide handy functions to build URLs."""

    def __init__(self, app: "Application"):
        self.app = app

    def url(self, path: str = "", query_params: dict = {}) -> str:
        """Generates a fully qualified url to the given path. If no path is given this will return
        the base url domain. A query parameters dictionary can be provided to add query parameters
        to the url."""
        # ensure that no slash is prefixing the relative path
        relative_path = path.lstrip("/")

        # add query params if any
        relative_path = add_query_params(relative_path, query_params)

        # fully qualify the url
        return join(config("application.app_url"), relative_path)

    def asset(self, alias: str, filename: str) -> str:
        """Generates a fully qualified URL for the given asset using the given disk
        Example:
            asset("local", "avatar.jpg") (take first pat)
            asset("s3.private", "doc.pdf") (when multiple paths are specified for the disk)
        """
        disks = config("filesystem.disks")
        # ensure that no slash is prefixing the relative filename path
        filename = filename.lstrip("/")
        if "." in alias:
            alias = alias.split(".")
            location = disks[alias[0]]["path"][alias[1]]
        else:
            location = disks[alias]["path"]
            # take first path if no path specified
            if isinstance(location, dict):
                location = list(location.values())[0]
        return join(location, filename)

    def route(
        self,
        name: str,
        params: dict = {},
        absolute: bool = True,
        query_params: dict = {},
    ) -> str:
        """Generates a fully qualified URL to the given route name.
        Example:
            route("users.home") : http://masonite.app/dashboard/
            route("users.profile", {"id": 1}) : http://masonite.app/users/1/profile/
            route("users.profile", {"id": 1}, query_params={"section": "infos"}) :
                http://masonite.app/users/1/profile/?section=info
            route("users.profile", {"id": 1}, absolute=False) : /users/1/profile/
        """

        relative_url = self.app.make("router").route(name, params, query_params)
        if absolute:
            return self.url(relative_url)
        else:
            return relative_url



import hashlib

from ..utils.filesystem import get_extension


class File:
    def __init__(self, content, filename=None):
        self.content = content
        self.filename = filename

    def path(self):
        pass

    def extension(self):
        return get_extension(self.filename)

    def name(self):
        return self.filename

    def stream(self):
        return self.content

    def hash_path_name(self):
        return f"{self.hash_name()}{self.extension()}"

    def hash_name(self):
        return hashlib.sha1(bytes(self.name(), "utf-8")).hexdigest()

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name()})"



import requests
from ..Recipient import Recipient


class MailgunDriver:
    def __init__(self, application):
        self.application = application
        self.options = {}
        self.content_type = None

    def set_options(self, options):
        self.options = options
        return self

    def get_mime_message(self):
        data = {
            "from": self.options.get("from"),
            "to": Recipient(self.options.get("to")).header(),
            "subject": self.options.get("subject"),
            "h:Reply-To": self.options.get("reply_to"),
            "html": self.options.get("html_content"),
            "text": self.options.get("text_content"),
        }

        if self.options.get("cc"):
            data.update({"cc", self.options.get("cc")})
        if self.options.get("bcc"):
            data.update({"bcc", self.options.get("bcc")})
        if self.options.get("priority"):
            data.update({"h:X-Priority", self.options.get("priority")})
        if self.options.get("headers"):
            for header, value in self.options.get("headers").items():
                data.update({f"h:{header}", value})

        return data

    def get_attachments(self):
        files = []
        for attachment in self.options.get("attachments", []):
            files.append(("attachment", open(attachment.path, "rb")))

        return files

    def send(self):
        domain = self.options["domain"]
        region = self.options.get("region", "us")
        secret = self.options["secret"]
        attachments = self.get_attachments()

        BASE_URL_BY_REGION = {
            "us": "https://api.mailgun.net/v3",
            "eu": "https://api.eu.mailgun.net/v3",
        }

        endpoint = BASE_URL_BY_REGION.get(region.lower())

        return requests.post(
            f"{endpoint}/{domain}/messages",
            auth=("api", secret),
            data=self.get_mime_message(),
            files=attachments,
        )



from .Command import Command


class DownCommand(Command):
    """
    Puts the server in a maintenance state.

    down
    """

    def handle(self):
        with open(".down", "w+"):
            pass

        self.info("Server is down for maintenance !")



from ...controllers import Controller
from ...request import Request
from ...response import Response
from ...authentication import Auth
from ..facades import Api


class AuthenticationController(Controller):
    def auth(self, auth: Auth, request: Request, response: Response):
        user = auth.attempt(request.input("username"), request.input("password"))

        if user:
            return {"data": user.generate_jwt()}

        return response.json(
            {"message": "Could not find username or password"}, status="403"
        )

    def reauth(self, request: Request, response: Response):
        user = Api.attempt_by_token(request.input("token"))

        if user:
            return {"data": user.generate_jwt()}

        return response.json(
            {"message": "Could not reauthenticate based on given token."}, status="403"
        )



from ..views import View
from .Provider import Provider
from ..helpers import UrlsHelper, MixHelper, optional
from ..configuration import config


class ViewProvider(Provider):
    def __init__(self, app):
        self.application = app

    def register(self):
        view = View(self.application)
        view.add_location(self.application.make("views.location"))

        self.application.bind("url", UrlsHelper(self.application))
        urls_helper = self.application.make("url")

        view.share(
            {
                "asset": urls_helper.asset,
                "mix": MixHelper(self.application).url,
                "url": urls_helper.url,
                "route": urls_helper.route,
                "config": config,
                "exists": view.exists,
                "optional": optional,
            }
        )

        self.application.bind("view", view)

    def boot(self):
        pass



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..foundation import Application
    from ..request import Request
    from ..response import Response


class Cors:
    """Service to handle Cross-Origin Resource Sharing operations."""

    def __init__(self, application: "Application", options: dict = {}):
        self.application = application
        self.options = {}
        self.allow_all_headers = False
        self.allow_all_methods = False
        self.allow_all_origins = False

        self.set_options(options)

    def set_options(self, options: dict) -> "Cors":

        # normalize options
        allowed_headers = list(
            map(lambda h: h.lower(), options.get("allowed_headers", []))
        )
        allowed_methods = list(
            map(lambda h: h.upper(), options.get("allowed_methods", []))
        )
        self.allow_all_headers = "*" in allowed_headers
        self.allow_all_methods = "*" in allowed_methods
        self.allow_all_origins = "*" in options.get("allowed_origins", [])

        self.options = {
            **options,
            "allowed_headers": allowed_headers,
            "allowed_methods": allowed_methods,
        }
        return self

    def is_single_origin_allowed(self) -> bool:
        if self.allow_all_origins:
            return False
        return len(self.options.get("allowed_origins", [])) == 1

    def is_preflight(self, request: "Request") -> bool:
        """Check if given request is a preflight request."""
        return request.get_request_method() == "OPTIONS" and request.header(
            "Access-Control-Request-Method"
        )

    def is_cors(self, request: "Request") -> bool:
        """Check if given request is CORS request by inspecting 'Origin' header."""
        return request.header("Origin")

    def is_allowed(self, request: "Request") -> bool:
        """Check if request 'Origin' is allowed regarding CORS config."""
        if self.allow_all_origins:
            return True

        origin = request.header("Origin")

        if origin in self.options.get("allowed_origins", []):
            return True

        return False

    def matches(self, request: "Request") -> bool:
        """Check if given request matches paths defined in CORS options."""
        pass

    def build_preflight_response(
        self, request: "Request", response: "Response"
    ) -> "Response":
        """Build a preflight response for the given request."""
        response.status(204)

        response = self.set_allowed_origin(request, response)

        if response.header("Access-Control-Allow-Origin"):
            response = self.set_allowed_credentials(response)
            response = self.set_allowed_methods(request, response)
            response = self.set_allowed_headers(request, response)
            response = self.set_max_age(response)

        return response

    def set_allowed_origin(
        self, request: "Request", response: "Response"
    ) -> "Response":
        if self.allow_all_origins and not self.options.get("supports_credentials"):
            response.header("Access-Control-Allow-Origin", "*")
        elif self.is_single_origin_allowed():
            response.header(
                "Access-Control-Allow-Origin", self.options.get("allowed_origins")[0]
            )
        else:
            if self.is_cors(request) and self.is_allowed(request):
                response.header("Access-Control-Allow-Origin", request.header("Origin"))

            self.update_vary_header(response, "Origin")
        return response

    def update_vary_header(self, response: "Response", value: str) -> "Response":
        if not response.header("Vary"):
            response.header("Vary", value)
        else:
            vary_header = response.header("Vary").split(", ")
            if value not in vary_header:
                vary_header.append(value)
                response.header("Vary", ", ".join(vary_header))
        return response

    def set_allowed_headers(
        self, request: "Request", response: "Response"
    ) -> "Response":
        if self.allow_all_headers:
            allowed_headers = request.header("Access-Control-Request-Headers")
            response = self.update_vary_header(
                response, "Access-Control-Request-Headers"
            )
        else:
            allowed_headers = ", ".join(self.options.get("allowed_headers", []))
        response.header("Access-Control-Allow-Headers", allowed_headers)
        return response

    def set_allowed_methods(
        self, request: "Request", response: "Response"
    ) -> "Response":
        if self.allow_all_methods:
            allowed_methods = request.header("Access-Control-Request-Method").upper()
            response = self.update_vary_header(
                response, "Access-Control-Request-Method"
            )
        else:
            allowed_methods = ", ".join(self.options.get("allowed_methods", []))
        response.header("Access-Control-Allow-Methods", allowed_methods)
        return response

    def set_max_age(self, response: "Response") -> "Response":
        max_age = self.options.get("max_age", None)
        # Here max_age could be set to 0s, different from None
        if max_age is not None:
            response.header("Access-Control-Max-Age", str(max_age))
        return response

    def set_exposed_headers(self, response: "Response") -> "Response":
        exposed_headers = self.options.get("exposed_headers", [])
        if exposed_headers:
            response.header("Access-Control-Expose-Headers", ", ".join(exposed_headers))
        return response

    def set_allowed_credentials(self, response: "Response") -> "Response":
        if self.options.get("supports_credentials", False):
            response.header("Access-Control-Allow-Credentials", "true")
        return response

    def add_actual_request_headers(
        self, request: "Request", response: "Response"
    ) -> "Response":
        """Add request headers to response."""
        response = self.set_allowed_origin(request, response)
        if response.header("Access-Control-Allow-Origin"):
            response = self.set_allowed_credentials(response)
            response = self.set_exposed_headers(response)
        return response



class Mail:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {}

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config.get(self.driver_config.get("default"), {})

        return self.driver_config.get(driver, {})

    def mailable(self, mailable):
        self.options = mailable.set_application(self.application).build().get_options()
        return self

    def send(self, driver=None):
        selected_driver = driver or self.options.get("driver", None)
        config_options = self.get_config_options(selected_driver)
        # if an option has already been defined in a mailable use it
        if self.options.get("from"):
            config_options.pop("from", None)
        self.options.update(config_options)
        return self.get_driver(selected_driver).set_options(self.options).send()



from urllib import parse
from ..utils.collections import flatten
from ..exceptions import RouteNotFoundException, MethodNotAllowedException


class Router:

    http_methods = ["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "OPTIONS"]

    def __init__(self, *routes, module_location=None):
        self.routes = flatten(routes)

    def find(self, path, request_method, subdomain=None):
        from .HTTPRoute import HTTPRoute

        for route in self.routes:
            if route.match(path, request_method, subdomain=subdomain):
                return route

        # we did not find a route matching the given path and method.
        # we will try to find a route matching other methods
        other_methods = [
            method for method in self.http_methods if method != request_method
        ]
        matched_methods = []
        for other_method in other_methods:
            for route in self.routes:
                if route.match(path, other_method, subdomain=subdomain):
                    matched_methods.append(other_method)
                    break

        # we really did not find a route
        if not matched_methods:
            return None

        # if alternative methods have been found, check if current request method is OPTIONS
        # to build a proper response else build a method not allowed response
        if request_method == "OPTIONS":
            def preflight_response(app):
                return (
                    app.make("response")
                    .with_headers({"Allow": ", ".join(matched_methods)})
                    .status(204)
                )

            preflight_route = HTTPRoute(path, request_method=["options"])
            preflight_route.get_response = preflight_response
            return preflight_route
        else:
            raise MethodNotAllowedException(matched_methods, request_method)

    def matches(self, path):
        for route in self.routes:
            if route.matches(path):
                return route

    def find_by_name(self, name):
        for route in self.routes:
            if route.match_name(name):
                return route

    def route(self, name: str, parameters: dict = {}, query_params: dict = {}) -> str:
        """Return URL string from given route name and parameters."""
        route = self.find_by_name(name)
        if route:
            return route.to_url(parameters, query_params)
        raise RouteNotFoundException(f"Could not find route with the name '{name}'")

    def set_controller_locations(self, location):
        self.controller_locations = location
        return self

    def add(self, *routes):
        self.routes.append(*routes)
        self.routes = flatten(self.routes)

    def set(self, *routes):
        self.routes = []
        self.routes.append(*routes)
        self.routes = flatten(self.routes)

    @classmethod
    def compile_to_url(cls, uncompiled_route, params={}, query_params={}):
        """Compile the route url into a usable url: converts /url/@id into /url/1.
        Used for redirection

        Arguments:
            route {string} -- An uncompiled route like (/dashboard/@user:string/@id:int)
        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route (default: {{}})
            query_params {dict} -- Dictionary of query parameters to pass to the route (default: {{}})
        Returns:
            string -- Returns a compiled string (/dashboard/joseph/1)
        """
        if "http" in uncompiled_route:
            return uncompiled_route

        # Split the url into a list
        split_url = uncompiled_route.split("/")

        # Start beginning of the new compiled url
        compiled_url = "/"

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if "@" in url:
                    url = url.replace("@", "").split(":")[0]
                    if isinstance(params, dict):
                        compiled_url += str(params[url]) + "/"
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + "/"
                elif "?" in url:
                    url = url.replace("?", "").split(":")[0]
                    if isinstance(params, dict):
                        compiled_url += str(params.get(url, "/")) + "/"
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + "/"
                else:
                    compiled_url += url + "/"

        compiled_url = compiled_url.replace("//", "")
        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith("/") and not uncompiled_route.endswith("/"):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if "//" in compiled_url:
            compiled_url = compiled_url.replace("//", "/")

        # Add eventual query parameters
        if query_params:
            compiled_url += "?" + parse.urlencode(query_params)

        return compiled_url



from ...providers import Provider
from ...utils.structures import load
from ..drivers import (
    BroadcastDriver,
    DatabaseDriver,
    MailDriver,
    SlackDriver,
    VonageDriver,
)
from ...configuration import config

from ..NotificationManager import NotificationManager
from ..MockNotification import MockNotification
from ..commands import MakeNotificationCommand, NotificationTableCommand


class NotificationProvider(Provider):
    """Notifications Provider"""

    def __init__(self, application):
        self.application = application

    def register(self):
        notification_manager = NotificationManager(self.application).set_configuration(
            config("notification")
        )
        notification_manager.add_driver("mail", MailDriver(self.application))
        notification_manager.add_driver("vonage", VonageDriver(self.application))
        notification_manager.add_driver("slack", SlackDriver(self.application))
        notification_manager.add_driver("database", DatabaseDriver(self.application))
        notification_manager.add_driver("broadcast", BroadcastDriver(self.application))

        self.application.bind("notification", notification_manager)
        self.application.bind("mock.notification", MockNotification)
        self.application.make("commands").add(
            MakeNotificationCommand(self.application),
            NotificationTableCommand(),
        )

    def boot(self):
        pass



"""Class modelling a Slack message."""
import json


class SlackMessage:
    WEBHOOK_MODE = 1
    API_MODE = 2

    def __init__(self):
        self._text = ""
        self._username = "masonite-bot"
        self._icon_emoji = ""
        self._icon_url = ""
        self._text = ""
        self._mrkdwn = True
        self._as_current_user = False
        self._reply_broadcast = False
        # Indicates if channel names and usernames should be linked.
        self._link_names = False
        # Indicates if you want a preview of links inlined in the message.
        self._unfurl_links = False
        # Indicates if you want a preview of links to media inlined in the message.
        self._unfurl_media = False
        self._blocks = []

        self._token = ""
        self._webhook = ""
        self._mode = None

    def from_(self, username, icon=None, url=None):
        """Set a custom username and optional emoji icon for the Slack message."""
        self._username = username
        if icon:
            self._icon_emoji = icon
        elif url:
            self._icon_url = url
        return self

    def to(self, to):
        """Specifies the channel to send the message to. It can be a list or single
        element. It can be either a channel ID or a channel name (with #), if it's
        a channel name the channel ID will be resolved later.
        """
        self._to = to
        return self

    def text(self, text):
        """Specifies the text to be sent in the message.

        Arguments:
            text {string} -- The text to show in the message.

        Returns:
            self
        """
        self._text = text
        return self

    def link_names(self):
        """Find and link channel names and usernames in message."""
        self._link_names = True
        return self

    def unfurl_links(self):
        """Whether the message should unfurl any links.

        Unfurling is when it shows a bigger part of the message after the text is sent
        like when pasting a link and it showing the header images.

        Returns:
            self
        """
        self._unfurl_links = True
        self._unfurl_media = True
        return self

    def without_markdown(self):
        """Specifies whether the message should explicitly not honor markdown text.

        Returns:
            self
        """
        self._mrkdwn = False
        return self

    def can_reply(self):
        """Whether the message should be ably to be replied back to.

        Returns:
            self
        """
        self._reply_broadcast = True
        return self

    def build(self, *args, **kwargs):
        return self

    def get_options(self):
        options = {
            "text": self._text,
            # optional
            "link_names": self._link_names,
            "unfurl_links": self._unfurl_links,
            "unfurl_media": self._unfurl_media,
            "username": self._username,
            "as_user": self._as_current_user,
            "icon_emoji": self._icon_emoji,
            "icon_url": self._icon_url,
            "mrkdwn": self._mrkdwn,
            "reply_broadcast": self._reply_broadcast,
            "blocks": json.dumps([block._resolve() for block in self._blocks]),
        }
        if self._mode == self.API_MODE:
            options.update({"channel": self._to, "token": self._token})
        return options

    def token(self, token):
        """[API_MODE only] Specifies the token to use for Slack authentication.

        Arguments:
            token {string} -- The Slack authentication token.

        Returns:
            self
        """
        self._token = token
        return self

    def as_current_user(self):
        """[API_MODE only] Send message as the currently authenticated user.

        Returns:
            self
        """
        self._as_current_user = True
        return self

    def webhook(self, webhook):
        """[WEBHOOK_MODE only] Specifies the webhook to use to send the message and authenticate
        to Slack.

        Arguments:
            webhook {string} -- Slack configured webhook url.

        Returns:
            self
        """
        self._webhook = webhook
        return self

    def block(self, block_instance):
        try:
            from slackblocks.blocks import Block
        except ImportError:
            raise ModuleNotFoundError(
                "Could not find the 'slackblocks' library. Run 'pip install slackblocks' to fix this."
            )

        if not isinstance(block_instance, Block):
            raise Exception("Blocks should be imported from 'slackblocks' package.")
        self._blocks.append(block_instance)
        return self

    def mode(self, mode):
        self._mode = mode
        return self



def validate(*rules, redirect=None, back=None):
    def decorator(func, rules=rules):
        def wrapper(*args, **kwargs):
            from wsgi import container

            request = container.make("Request")
            response = container.make("Response")
            errors = request.validate(*rules)
            if errors:
                if redirect:
                    return response.redirect(redirect).with_errors(errors).with_input()
                if back:
                    return response.back().with_errors(errors).with_input()
                return errors
            else:
                return container.resolve(func)

        return wrapper

    return decorator



from .Input import Input
from urllib.parse import parse_qs
import re
import json
import cgi
import re
from ..utils.structures import data_get
from ..filesystem import UploadedFile


class InputBag:
    def __init__(self):
        self.query_string = {}
        self.post_data = {}
        self.environ = {}

    def load(self, environ):
        self.environ = environ
        self.query_string = {}
        self.post_data = {}
        self.parse(environ)
        return self

    def parse(self, environ):
        if "QUERY_STRING" in environ:
            self.query_string = self.query_parse(environ["QUERY_STRING"])

        if "wsgi.input" in environ:
            if "application/json" in environ.get("CONTENT_TYPE", ""):
                try:
                    request_body_size = int(environ.get("CONTENT_LENGTH", 0))
                except (ValueError):
                    request_body_size = 0

                request_body = environ["wsgi.input"].read(request_body_size)

                if isinstance(request_body, bytes):
                    request_body = request_body.decode("utf-8")

                json_payload = json.loads(request_body or "{}")
                if isinstance(json_payload, list):
                    pass
                else:
                    for name, value in json.loads(request_body or "{}").items():
                        self.post_data.update({name: value})

            elif "application/x-www-form-urlencoded" in environ.get("CONTENT_TYPE", ""):
                try:
                    request_body_size = int(environ.get("CONTENT_LENGTH", 0))
                except (ValueError):
                    request_body_size = 0

                request_body = environ["wsgi.input"].read(request_body_size)
                parsed_request_body = parse_qs(bytes(request_body).decode("utf-8"))

                self.post_data = self.parse_dict(parsed_request_body)

            elif "multipart/form-data" in environ.get("CONTENT_TYPE", ""):
                fields = cgi.FieldStorage(
                    fp=environ["wsgi.input"],
                    environ=environ,
                    keep_blank_values=1,
                )

                for name in fields:
                    value = fields.getvalue(name)
                    if isinstance(value, list):
                        files = []
                        k = 0
                        for item in value:
                            files.append(
                                UploadedFile(fields[name][k].filename, value[k])
                            )
                            k += 1
                        self.post_data.update({name: files})
                    elif isinstance(value, bytes):
                        self.post_data.update(
                            {name: [UploadedFile(fields[name].filename, value)]}
                        )
                    else:
                        self.post_data.update({name: value})

                self.post_data = self.parse_dict(self.post_data)
            else:
                try:
                    request_body_size = int(environ.get("CONTENT_LENGTH", 0))
                except (ValueError):
                    request_body_size = 0

                request_body = environ["wsgi.input"].read(request_body_size)
                if request_body:
                    self.post_data.update(
                        json.loads(bytes(request_body).decode("utf-8"))
                    )

    def get(self, name, default=None, clean=True, quote=True):
        if isinstance(name, str) and name.endswith("[]"):
            default = []

        input = data_get(self.all(), name, default)

        if isinstance(input, (str,)):
            return input
        if isinstance(input, list) and len(input) == 1:
            if name.endswith("[]"):
                return input

            return input[0]
        elif isinstance(input, (dict,)):
            rendered = {}
            for key, inp in input.items():
                if hasattr(inp, "value"):
                    inp = inp.value
                rendered.update({key: inp})
            return rendered
        elif hasattr(input, "value"):
            if isinstance(input.value, list) and len(input.value) == 1:
                """
                return input.value[0]

                    This line is converting request input list to object if the input is a list having only one item

                    Problem:
                        1. This will make validations and request input confusing as a developer is sending array where as
                            they will get dict in controller, this is actually a bug rather than a feature
                        2. In case of nested validations, this will make the validation to fail
                            Example:
                                input => {
                                    "test": [
                                        {
                                            "foo": "bar"
                                        }
                                    ],
                                    "test_1": {
                                        "foo": "bar"
                                    }
                                }
                                validation => validate.required(["test.*.foo"])

                            In above example `test` and `test_1` are not same but this code `input.value[0]` will make them treat as same

                    Solution:
                        return the input value without removing anything...
                """
                return input.value
            elif isinstance(input.value, dict):
                return input.value
            return input.value

        return input

    def has(self, *names):
        return all((name in self.all()) for name in names)

    def all(self):
        all = {}
        qs = self.query_string
        if isinstance(qs, list):
            qs = {str(i): v for i, v in enumerate(qs)}

        all.update(qs)
        all.update(self.post_data)
        return all

    def all_as_values(self, internal_variables=False):
        all = self.all()
        new = {}
        for name, input in all.items():
            if not internal_variables:
                if name.startswith("__"):
                    continue
            new.update({name: self.get(name)})

        return new

    def only(self, *args):
        all = self.all()
        new = {}
        for name, input in all.items():
            if name not in args:
                continue
            new.update({name: self.get(name)})

        return new

    def query_parse(self, query_string):
        return self.parse_dict(parse_qs(query_string))

    def parse_dict(self, dictionary):
        d = {}
        for name, value in dictionary.items():
            if name.endswith("[]"):
                d.update({name: value})
            else:
                regex_match = re.match(r"(?P<name>[^\[]+)\[(?P<value>[^\]]+)\]", name)

                if regex_match:
                    gd = regex_match.groupdict()
                    if isinstance(value, Input):
                        d.setdefault(gd["name"], {})[gd["value"]] = value
                    else:
                        d.setdefault(gd["name"], {})[gd["value"]] = value[0]
                else:
                    if isinstance(value, (list, tuple)):
                        d.update({name: value[0]})
                    else:
                        d.update({name: value})

        new_dict = {}
        # Further filter the dictionary
        for name, value in d.items():
            if "[]" in name:
                new_name = name.replace("[]", "")
                regex_match = re.match(
                    r"(?P<name>[^\[]+)*\[(?P<value>[^\]]+)\]", new_name
                )
                if regex_match:
                    new_dict.setdefault(regex_match["name"], []).append(
                        {regex_match["value"]: value}
                    )
                else:
                    new_dict.update({name: value})
            else:
                new_dict.update({name: value})

        return new_dict

    def add_post(self, key, value):
        self.post_data.update({key: value})
        return value



"""Database notification driver."""
import json

from .BaseDriver import BaseDriver


class DatabaseDriver(BaseDriver):
    def __init__(self, application):
        self.application = application
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def get_builder(self):
        return (
            self.application.make("builder")
            .on(self.options.get("connection"))
            .table(self.options.get("table"))
        )

    def send(self, notifiable, notification):
        """Used to send the email and run the logic for sending emails."""
        data = self.build(notifiable, notification)
        return self.get_builder().new().create(data)

    def build(self, notifiable, notification):
        """Build an array payload for the DatabaseNotification Model."""
        return {
            "id": str(notification.id),
            "type": notification.type(),
            "notifiable_id": notifiable.id,
            "notifiable_type": notifiable.get_table_name(),
            "data": json.dumps(self.get_data("database", notifiable, notification)),
            "read_at": None,
        }



from ...providers import Provider
from ..PresetsCapsule import PresetsCapsule
from .. import Tailwind, Vue, React, Bootstrap


class PresetsProvider(Provider):
    """
    @M5
    This provider will be used in Masonite 5. For now, this code is added
    into the FrameworkProvider to allow projects without PresetsProvider to
    work."""

    def __init__(self, app):
        self.application = app

    def register(self):
        presets = PresetsCapsule()
        presets.add(Bootstrap())
        presets.add(Tailwind())
        presets.add(Vue())
        presets.add(React())
        self.application.bind("presets", presets)

    def boot(self):
        pass



"""__class__ validation"""

from masonite.validation import BaseValidation


class __class__(BaseValidation):
    """__class__ validation class"""

    def passes(self, attribute, key, dictionary):
        """The passing criteria for this rule.

        This should return a True boolean value.

        Arguments:
            attribute {mixed} -- The value found within the dictionary
            key {string} -- The key in the dictionary being searched for.
                            This key may or may not exist in the dictionary.
            dictionary {dict} -- The dictionary being searched

        Returns:
            bool
        """
        return attribute

    def message(self, key):
        """A message to show when this rule fails

        Arguments:
            key {string} -- The key used to search the dictionary

        Returns:
            string
        """
        return f"{key} is required"

    def negated_message(self, key):
        """A message to show when this rule is negated using a negation rule like 'isnt()'

        For example if you have a message that says 'this is required' you may have a negated statement
        that says 'this is not required'.

        Arguments:
            key {string} -- The key used to search the dictionary

        Returns:
            string
        """
        return "{key} is not required"



"""New Controller Command."""
import inflection
import os
from ..utils.location import controllers_path
from ..utils.filesystem import get_module_dir, render_stub_file
from .Command import Command


class MakeControllerCommand(Command):
    """
    Creates a new controller class.

    controller
        {name : Name of the controller}
        {--r|--resource : Create a "resource" controller with the usual CRUD methods}
        {--a|--api : Create an "api" controller with the usual CRUD methods}
        {--f|force=? : Force overriding file if already exists}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):

        full_path = self.argument("name")

        if "/" in full_path:
            name = inflection.camelize(os.path.basename(full_path))
            parent_directory = os.path.dirname(full_path)
        else:
            name = inflection.camelize(full_path)
            parent_directory = ""

        folder_controller = controllers_path(parent_directory)
        os.makedirs(folder_controller, exist_ok=True)

        if not name.endswith("Controller"):
            name += "Controller"

        # create a resource controller if required
        if self.option("resource"):
            stub_path = self.get_resource_controller_path()
        elif self.option("api"):
            stub_path = self.get_api_controller_path()
        else:
            stub_path = self.get_basic_controller_path()

        content = render_stub_file(stub_path, name)

        filename = f"{name}.py"
        full_path_with_name = os.path.join(
            controllers_path(folder_controller), filename
        )

        if os.path.exists(full_path_with_name) and not self.option("force"):
            self.warning(
                f"{filename} already exists! Run the command with -f (force) to override."
            )
            return -1

        with open(full_path_with_name, "w") as f:
            f.write(content)

        # add class to __init__.py
        with open(
            os.path.join(controllers_path(folder_controller), "__init__.py"), "a"
        ) as f:
            f.write(f"from .{name} import {name}\n")

        file_created = os.path.join(parent_directory, filename)
        self.info(
            f"Controller Created ({controllers_path(file_created, absolute=False)})"
        )

    def get_basic_controller_path(self):
        return os.path.join(
            get_module_dir(__file__), "../stubs/controllers/Controller.py"
        )

    def get_resource_controller_path(self):
        return os.path.join(
            get_module_dir(__file__), "../stubs/controllers/ResourceController.py"
        )

    def get_api_controller_path(self):
        return os.path.join(
            get_module_dir(__file__), "../stubs/controllers/APIController.py"
        )



"""String generators and helpers"""
import random
import string
from urllib import parse
from typing import Any


def random_string(length=4):
    """Generate a random string based on the given length.

    Keyword Arguments:
        length {int} -- The amount of the characters to generate (default: {4})

    Returns:
        string
    """
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(length)
    )


def modularize(file_path, suffix=".py"):
    """Transforms a file path to a dotted path. On UNIX paths contains / and on Windows \\.

    Keyword Arguments:
        file_path {str} -- A file path such app/controllers

    Returns:
        value {str} -- a dotted path such as app.controllers
    """
    # if the file had the .py extension remove it as it's not needed for a module
    return removesuffix(file_path.replace("/", ".").replace("\\", "."), suffix)


def as_filepath(dotted_path):
    """Inverse of modularize, transforms a dotted path to a file path (with /).

    Keyword Arguments:
        dotted_path {str} -- A dotted path such app.controllers

    Returns:
        value {str} -- a file path such as app/controllers
    """
    return dotted_path.replace(".", "/")


def removeprefix(string, prefix):
    """Implementation of str.removeprefix() function available for Python versions lower than 3.9."""
    if string.startswith(prefix):
        return string[len(prefix) :]
    else:
        return string


def removesuffix(string, suffix):
    """Implementation of str.removesuffix() function available for Python versions lower than 3.9."""
    if suffix and string.endswith(suffix):
        return string[: -len(suffix)]
    else:
        return string


def match(string: str, ref_string: str) -> str:
    """Check if a given string matches a reference string. Wildcard '*' can be used at start, end
    or middle of the string."""
    if ref_string.startswith("*"):
        ref_string = ref_string.replace("*", "")
        return string.endswith(ref_string)
    elif ref_string.endswith("*"):
        ref_string = ref_string.replace("*", "")
        return string.startswith(ref_string)
    elif "*" in ref_string:
        split_search = ref_string.split("*")
        return string.startswith(split_search[0]) and string.endswith(split_search[1])
    else:
        return ref_string == string


def add_query_params(url: str, query_params: dict) -> str:
    """Add query params dict to a given url (which can already contain some query parameters)."""
    path_result = parse.urlsplit(url)

    base_url = path_result.path

    # parse existing query parameters if any
    existing_query_params = dict(parse.parse_qsl(path_result.query))
    all_query_params = {**existing_query_params, **query_params}

    # add query parameters to url if any
    if all_query_params:
        base_url += "?" + parse.urlencode(all_query_params)

    return base_url


def get_controller_name(controller: "str|Any") -> str:
    """Get a controller string name from a controller argument used in routes."""
    # controller is a class or class.method
    if hasattr(controller, "__qualname__"):
        if "." in controller.__qualname__:
            controller_str = controller.__qualname__.replace(".", "@")
        else:
            controller_str = f"{controller.__qualname__}@__call__"
    # controller is an instance, so the method will automatically be __call__
    elif not isinstance(controller, str):
        controller_str = f"{controller.__class__.__qualname__}@__call__"
    # controller is anything else: "Controller@method"
    else:
        controller_str = str(controller)
    return controller_str



from masonite.mail import Mailable
from masonite.configuration import config


class ResetPassword(Mailable):
    def __init__(self, token=None):
        super().__init__()
        self.token = token

    def build(self):
        return (
            self.subject("Reset Password")
            .from_(config("mail.from_address"))
            .view("auth.mailables.reset_password", {"token": self.token})
        )



class PresenceChannel:
    def __init__(self, name):
        if not name.startswith("presence-"):
            name = "presence-" + name

        self.name = name

    def authorized(self, application):
        return bool(application.make("request").user())



"""New Listener Command."""
from cleo import Command
import inflection
import os

from ...utils.filesystem import make_directory, get_module_dir, render_stub_file
from ...utils.str import as_filepath
from ...utils.location import base_path


class MakeListenerCommand(Command):
    """
    Creates a new listener class.

    listener
        {name : Name of the listener}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        content = render_stub_file(self.get_path(), name)

        relative_filename = os.path.join(
            as_filepath(self.app.make("listeners.location")), f"{name}.py"
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)

        with open(filepath, "w") as f:
            f.write(content)
        self.info(f"Listener Created ({relative_filename})")

    def get_path(self):
        return os.path.join(get_module_dir(__file__), "../../stubs/events/Listener.py")



from ...providers import Provider
from .. import RateLimiter


class RateProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        rate_limiter = RateLimiter(self.application)
        self.application.bind("rate", rate_limiter)

    def boot(self):
        pass



""" A ScheduleProvider Service Provider """
from ...providers import Provider

from ..commands import MakeTaskCommand, ScheduleRunCommand
from ..TaskHandler import TaskHandler


class ScheduleProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        self.application.make("commands").add(
            MakeTaskCommand(self.application), ScheduleRunCommand(self.application)
        )

        self.application.bind("scheduler", TaskHandler(self.application))

    def boot(self):
        pass



class Recipient:
    def __init__(self, recipient):
        if isinstance(recipient, (list, tuple)):
            recipient = ",".join(recipient)
        self.recipient = recipient

    def header(self):
        headers = []
        for address in self.recipient.split(","):

            if "<" in address:
                headers.append(address)
                continue

            if address.strip():
                headers.append(f"<{address.strip()}>")

        return ", ".join(headers)



from .NotificationManager import NotificationManager
from .AnonymousNotifiable import AnonymousNotifiable
from .Notification import Notification


class NotificationWithAsserts(Notification):
    def assertSentVia(self, *drivers):
        sent_via = self.via(self.notifiable)
        for driver in drivers:
            assert (
                driver in sent_via
            ), f"notification sent via {sent_via}, not {driver}."
        return self

    def assertEqual(self, value, reference):
        assert value == reference, "{value} not equal to {reference}."
        return self

    def assertNotEqual(self, value, reference):
        assert value != reference, "{value} equal to {reference}."
        return self

    def assertIn(self, value, container):
        assert value in container, "{value} not in {container}."
        return self

    @classmethod
    def patch(cls, target):
        for k in cls.__dict__:
            obj = getattr(cls, k)
            if not k.startswith("_") and callable(obj):
                setattr(target, k, obj)


class MockNotification(NotificationManager):
    def __init__(self, application, *args, **kwargs):
        super().__init__(application, *args, **kwargs)
        self.count = 0
        self.last_notifiable = None
        self.last_notification = None

    def send(
        self, notifiables, notification, drivers=[], dry=False, fail_silently=False
    ):
        _notifiables = []
        for notifiable in self._format_notifiables(notifiables):
            if isinstance(notifiable, AnonymousNotifiable):
                _notifiables.extend(notifiable._routes.values())
            else:
                _notifiables.append(notifiable)

        notification_key = notification.type()
        NotificationWithAsserts.patch(notification.__class__)
        for notifiable in _notifiables:
            notification.notifiable = notifiable  # for asserts
            old_notifs = self.sent_notifications.get(notifiable, {})
            old_notifs.update(
                {
                    notification_key: old_notifs.get(notification_key, [])
                    + [notification]
                }
            )
            self.sent_notifications.update({notifiable: old_notifs})
            self.count += 1
        self.last_notification = notification
        self.last_notifiable = notifiable
        return self

    def resetCount(self):
        """Reset sent notifications count."""
        self.count = 0
        self.sent_notifications = {}
        self.last_notifiable = None
        self.last_notification = None
        return self

    def assertNothingSent(self):
        assert self.count == 0, f"{self.count} notifications have been sent."
        return self

    def assertCount(self, count):
        assert (
            self.count == count
        ), f"{self.count} notifications have been sent, not {count}."
        return self

    def assertSentTo(
        self, notifiable, notification_class, callable_assert=None, count=None
    ):
        notification_key = notification_class.__name__
        notifiable_notifications = self.sent_notifications.get(notifiable, [])
        assert notification_key in notifiable_notifications
        if count:
            sent_count = len(notifiable_notifications.get(notification_key, []))
            assert (
                sent_count == count
            ), f"{notification_key} has been sent to {notifiable} {sent_count} times"
        if callable_assert:
            # assert last notification sent for this notifiable
            notification = notifiable_notifications.get(notification_key)[-1]
            assert callable_assert(notifiable, notification)
        return self

    def last(self):
        """Get last sent mocked notification if any."""
        return self.last_notification

    def assertLast(self, callable_assert):
        if not self.last_notifiable or not self.last_notification:
            raise AssertionError("No notification has been sent.")
        assert callable_assert(self.last_notifiable, self.last_notification)
        return self

    def assertNotSentTo(self, notifiable, notification_class):
        notification_key = notification_class.__name__
        notifiable_notifications = self.sent_notifications.get(notifiable, [])
        assert (
            notification_key not in notifiable_notifications
        ), f"{notification_key} has been sent to {notifiable}."
        return self



from ..facades import Gate


class AuthorizesRequest:
    """Request mixin to add permission handling to requests."""

    def authorize(self, permission: str, *args):
        """Check if the user performing the current request has the given permission."""
        return Gate.authorize(permission, *args)



"""Anonymous Notifiable mixin"""

from .Notifiable import Notifiable


class AnonymousNotifiable(Notifiable):
    """Anonymous notifiable allowing to send notification without having
    a notifiable entity.

    Usage:
        self.notification.route("sms", "+3346474764").send(WelcomeNotification())
    """

    def __init__(self, application=None):
        self.application = application
        self._routes = {}

    def route(self, driver, recipient):
        """Define which driver using to route the notification."""
        if driver == "database":
            raise ValueError(
                "The database driver does not support on-demand notifications."
            )
        self._routes[driver] = recipient
        return self

    def route_notification_for(self, driver):
        try:
            return self._routes[driver]
        except KeyError:
            raise ValueError(
                "Routing has not been defined for the driver {}".format(driver)
            )

    def send(self, notification, dry=False, fail_silently=False):
        """Send the given notification."""
        return self.application.make("notification").send(
            self, notification, self._routes, dry, fail_silently
        )



from .Provider import Provider
from whitenoise import WhiteNoise
import os


class WhitenoiseProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):

        response_handler = WhiteNoise(
            self.application.get_response_handler(),
            root=self.application.get_storage_path(),
            autorefresh=True,
        )

        for location, alias in (
            self.application.make("storage_capsule").get_storage_assets().items()
        ):
            response_handler.add_files(location, prefix=alias)

        self.application.set_response_handler(response_handler)

    def boot(self):
        return



from setuptools import setup
import os

meta = {}
with open(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "src/masonite", "__init__.py"
    ),
    "r",
) as f:
    exec(f.read(), meta)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="masonite",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=meta["__version__"],
    package_dir={"": "src"},
    description="The Masonite Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/masoniteframework/masonite",
    # Author details
    author="Joe Mancuso",
    author_email="joe@masoniteproject.com",
    # Choose your license
    license="MIT",
    # If your package should include things you specify in your MANIFEST.in file
    # Use this option if your package needs to include files that are not python files
    # like html templates or css files
    include_package_data=True,
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "inflection>=0.3<0.4",
        "exceptionite>=2.2<3.0",
        "pendulum>=2,<3",
        "jinja2<3.1.0",
        "cleo>=0.8.1,<0.9",
        "hupper>=1.10,<1.11",
        "bcrypt>=3.2,<3.3",
        "whitenoise>=5.2,<5.3",
        "python-dotenv>=0.15,<0.16",
        "masonite-orm>=2.14,<3",
        "hashids>=1.3,<1.4",
        "cryptography>=36<37",
        "tldextract>=2.2,<2.3",
        "hfilesize>=0.1",
        "dotty_dict>=1.3.0<1.40",
        "pyjwt>=2.3,<2.5",
        "pytest>=7<8",
        "werkzeug>=2<3",
        "watchdog>=2<3",
    ],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Environment :: Web Environment",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    # What does your project relate to?
    keywords="Masonite, MasoniteFramework, Python, ORM",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=[
        "masonite",
        "masonite.api",
        "masonite.api.authentication",
        "masonite.api.commands",
        "masonite.api.controllers",
        "masonite.api.facades",
        "masonite.api.guards",
        "masonite.api.middleware",
        "masonite.api.providers",
        "masonite.api.stubs",
        "masonite.api.stubs.routes",
        "masonite.auth",
        "masonite.authentication.guards",
        "masonite.authentication.models",
        "masonite.authentication",
        "masonite.authorization.models",
        "masonite.authorization",
        "masonite.broadcasting.controllers",
        "masonite.broadcasting.drivers",
        "masonite.broadcasting.providers",
        "masonite.broadcasting",
        "masonite.cache.drivers",
        "masonite.cache",
        "masonite.commands",
        "masonite.configuration.providers",
        "masonite.configuration",
        "masonite.container",
        "masonite.controllers",
        "masonite.cookies",
        "masonite.cors",
        "masonite.drivers.queue",
        "masonite.drivers.session",
        "masonite.drivers",
        "masonite.dumps",
        "masonite.environment",
        "masonite.essentials.helpers",
        "masonite.essentials.middleware",
        "masonite.essentials.providers",
        "masonite.essentials",
        "masonite.events.commands",
        "masonite.events.providers",
        "masonite.events",
        "masonite.exceptions",
        "masonite.exceptions.exceptionite",
        "masonite.exceptions.handlers",
        "masonite.facades",
        "masonite.filesystem.drivers",
        "masonite.filesystem.providers",
        "masonite.filesystem",
        "masonite.foundation",
        "masonite.hashing.drivers",
        "masonite.hashing",
        "masonite.headers",
        "masonite.helpers",
        "masonite.input",
        "masonite.loader",
        "masonite.mail.drivers",
        "masonite.mail",
        "masonite.middleware.route",
        "masonite.middleware",
        "masonite.notification.commands",
        "masonite.notification.drivers.vonage",
        "masonite.notification.drivers",
        "masonite.notification.providers",
        "masonite.notification",
        "masonite.orm",
        "masonite.orm.commands",
        "masonite.orm.providers",
        "masonite.packages.providers",
        "masonite.packages",
        "masonite.pipeline.tasks",
        "masonite.pipeline",
        "masonite.presets",
        "masonite.presets.providers",
        "masonite.providers",
        "masonite.queues",
        "masonite.request",
        "masonite.response",
        "masonite.rates",
        "masonite.rates.providers",
        "masonite.routes",
        "masonite.routes.commands",
        "masonite.scheduling.commands",
        "masonite.scheduling.providers",
        "masonite.scheduling",
        "masonite.sessions",
        "masonite.storage",
        "masonite.templates",
        "masonite.tests",
        "masonite.utils",
        "masonite.validation.commands",
        "masonite.validation.providers",
        "masonite.validation.resources",
        "masonite.validation",
        "masonite.views",
    ],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # $ pip install your-package[dev,test]
    extras_require={
        "test": [
            "coverage",
            "pytest",
            "redis",
            "boto3",
            "pusher",
            "pymemcache",
            "vonage",
            "slackblocks",
            "argon2-cffi",
            "pwnedapi",
        ],
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    ## package_data={
    ##     'sample': [],
    ## },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    ## data_files=[('my_data', ['data/data_file.txt'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": [
            "project = masonite.commands.Entry:application.run",
        ],
    },
)



import hashlib
from typing import TYPE_CHECKING

from attr import has

if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

from ..middleware import Middleware
from ...facades import RateLimiter
from ...rates.Limit import Limit
from ...exceptions import ThrottleRequestsException


class ThrottleRequestsMiddleware(Middleware):
    def before(
        self, request: "Request", response: "Response", limit_string: str
    ) -> "Request":
        response_callback = None
        if "/" in limit_string:
            limit = Limit.from_str(limit_string)
        else:
            limiter = RateLimiter.get_limiter(limit_string)
            if hasattr(limiter, "get_response"):
                response_callback = limiter.get_response
            limit = limiter.allow(request)

        if limit.is_unlimited():
            return request
        else:
            key = hashlib.md5(str(limit_string + limit.key).encode()).hexdigest()
            if RateLimiter.too_many_attempts(key, limit.max_attempts):
                headers = self.get_headers(key, limit.max_attempts, True)
                # raise exception or show custom response of limiter
                if response_callback:
                    return response_callback(request, response, headers)
                else:
                    raise ThrottleRequestsException(headers=headers)
            else:
                # delay rates are in minutes
                RateLimiter.hit(key, limit.delay * 60)

        # add headers to response
        response.with_headers(self.get_headers(key, limit.max_attempts))

        return request

    def after(
        self, request: "Request", response: "Response", limit_string: str
    ) -> "Request":
        return request

    def get_headers(self, key: str, max_attempts: int, limited: bool = False) -> dict:
        headers = {
            "X-RateLimit-Limit": max_attempts,
            "X-RateLimit-Remaining": RateLimiter.remaining(key, max_attempts),
        }
        if limited:
            headers.update(
                {
                    "Retry-After": RateLimiter.available_in(key),
                    "X-RateLimit-Reset": RateLimiter.available_at(key),
                }
            )
        return headers



from .. import Middleware
from markupsafe import Markup
from ...exceptions import InvalidCSRFToken
from hmac import compare_digest


class VerifyCsrfToken(Middleware):

    exempt = []

    def before(self, request, response):
        self.verify_token(request, self.get_token(request))

        token = self.create_token(request, response)

        request.app.make("view").share(
            {
                "csrf_field": Markup(
                    f"<input type='hidden' name='__token' value='{token}' />"
                ),
                "csrf_token": token,
            }
        )

        return request

    def after(self, request, response):
        return request

    def create_token(self, request, response):
        session = request.cookie("SESSID")
        response.cookie("csrf_token", session)
        return session

    def verify_token(self, request, token):
        if self.in_exempt(request):
            return True
        if request.is_not_safe() and not token:
            raise InvalidCSRFToken("Missing CSRF Token")
        if request.is_not_safe():
            if request.cookie("csrf_token") and (
                compare_digest(
                    request.cookie("csrf_token"),
                    token,
                )
                and compare_digest(token, request.cookie("SESSID"))
            ):
                return True
            raise InvalidCSRFToken("Invalid CSRF Token")
        return True

    def in_exempt(self, request):
        """Determine if the request has a URI that should pass through CSRF verification.

        Returns:
            bool
        """
        for route in self.exempt:
            if request.contains(route):
                return True

        return False

    def get_token(self, request):
        return (
            request.header("X-CSRF-TOKEN")
            or request.header("X-XSRF-TOKEN")
            or request.input("__token")
        )



import os
import shutil
import json

from ..utils.location import base_path, resources_path
from ..utils.filesystem import get_module_dir, make_full_directory


class Preset:
    """Base preset class containing business logic for adding a preset to the project."""

    key = ""
    packages = {}
    removed_packages = []

    def get_base_stubs_directory(self):
        return os.path.join(get_module_dir(__file__), "../stubs/presets/base/")

    def get_stubs_directory(self):
        return os.path.join(get_module_dir(__file__), f"../stubs/presets/{self.key}/")

    def get_base_template_path(self, template):
        return os.path.join(self.get_base_stubs_directory(), template)

    def get_template_path(self, template):
        return os.path.join(self.get_stubs_directory(), template)

    def update_webpack_mix(self):
        """Replace webpack.mix.js with the one from preset."""
        shutil.copyfile(
            self.get_template_path("webpack.mix.js"), base_path("webpack.mix.js")
        )

    def update_packages(self, dev=True):
        """Update the "package.json" file."""
        filepath = base_path("package.json")
        if not os.path.exists(filepath):
            return

        configuration_key = "devDependencies" if dev else "dependencies"

        with open(filepath, "r+") as f:
            packages = json.load(f)
            # add new packages
            packages[configuration_key].update(self.packages)
            # remove packages
            for package in self.removed_packages:
                packages[configuration_key].pop(package, None)

            f.seek(0)  # Rewind to beginning of file
            f.truncate()
            f.write(json.dumps(packages, sort_keys=True, indent=4))

    def update_css(self):
        """Create/Override an app.css file configured for the preset."""
        make_full_directory(resources_path("css"))
        shutil.copyfile(
            self.get_base_template_path("app.css"), resources_path("css/app.css")
        )

    def update_js(self):
        """Create/Override an app.js file configured for the preset."""
        make_full_directory(resources_path("js"))
        shutil.copyfile(self.get_template_path("app.js"), resources_path("js/app.js"))
        shutil.copyfile(
            self.get_base_template_path("bootstrap.js"),
            resources_path("js/bootstrap.js"),
        )

    def remove_node_modules(self):
        """Remove the installed Node modules."""
        for filename in ["package-lock.json", "yarn.lock"]:
            if os.path.exists(base_path(filename)):
                os.remove(filename)
        shutil.rmtree(base_path("node_modules"), ignore_errors=True)



class BaseDriver:
    def send(self, notifiable, notification):
        """Implements sending the notification to notifiables through
        this driver."""
        raise NotImplementedError(
            "send() method must be implemented for a notification driver."
        )

    def get_data(self, driver, notifiable, notification):
        """Get the data for the notification."""
        method_name = f"to_{driver}"
        try:
            method = getattr(notification, method_name)
        except AttributeError:
            raise NotImplementedError(
                f"Notification model should implement {method_name}() method."
            )
        else:
            return method(notifiable)



from masoniteorm.migrations import Migration


class CreateFailedJobsTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("failed_jobs") as table:
            table.increments("id")
            table.string("queue").nullable()
            table.string("connection").nullable()
            table.string("name").nullable()
            table.string("driver").nullable()
            table.binary("payload")
            table.text("exception").nullable()
            table.timestamp("failed_at").nullable()
            table.timestamp("created_at").nullable()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("failed_jobs")



"""Helpers to resolve absolute paths to the different app resources using a configured
location."""
from os.path import join, abspath

from .str import as_filepath


def _build_path(location_key, relative_path, absolute):
    from wsgi import application

    relative_dir = join(as_filepath(application.make(location_key)), relative_path)
    return abspath(relative_dir) if absolute else relative_dir


def base_path(relative_path=""):
    """Build the absolute path to the project root directory or build the absolute path to a
    given file relative to the project root directory."""
    return abspath(relative_path)


def views_path(relative_path="", absolute=True):
    """Build the absolute path to the project views directory or build the absolute path to a given
    file relative to the project views directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("views.location", relative_path, absolute)


def controllers_path(relative_path="", absolute=True):
    """Build the absolute path to the project controllers directory or build the absolute path to a given
    file relative to the project controllers directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("controllers.location", relative_path, absolute)


def mailables_path(relative_path="", absolute=True):
    """Build the absolute path to the project controllers directory or build the absolute path to a given
    file relative to the project controllers directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("mailables.location", relative_path, absolute)


def config_path(relative_path="", absolute=True):
    """Build the absolute path to the project configuration directory or build the absolute path to a given
    file relative to the project configuration directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("config.location", relative_path, absolute)


def migrations_path(relative_path="", absolute=True):
    """Build the absolute path to the project migrations directory or build the absolute path to a given
    file relative to the project migrations directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("migrations.location", relative_path, absolute)


def seeds_path(relative_path="", absolute=True):
    """Build the absolute path to the project seeds directory or build the absolute path to a given
    file relative to the project seeds directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("seeds.location", relative_path, absolute)


def jobs_path(relative_path="", absolute=True):
    """Build the absolute path to the project jobs directory or build the absolute path to a given
    file relative to the project jobs directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("jobs.location", relative_path, absolute)


def resources_path(relative_path="", absolute=True):
    """Build the absolute path to the project resources directory or build the absolute path to a given
    file relative to the project resources directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("resources.location", relative_path, absolute)


def models_path(relative_path="", absolute=True):
    """Build the absolute path to the project models directory or build the absolute path to a given
    file relative to the project models directory.

    The relative path can be returned instead by setting absolute=False."""
    return _build_path("models.location", relative_path, absolute)



from werkzeug.serving import run_simple

from .Command import Command


class ServeCommand(Command):
    """
    Run the Masonite server.

    serve
        {--p|port=8000 : Specify which port to run the server}
        {--b|host=127.0.0.1 : Specify which ip address to run the server}
        {--d|dont-reload : Make the server NOT automatically reload on file changes}
        {--i|reload-interval=1 : Number of seconds to wait to reload after changed are detected}
        {--l|live-reload : Make the server automatically refresh your web browser}
        {--t|threaded : Handle concurrent requests using threads}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        if self.option("live-reload"):
            try:
                from livereload import Server
            except ImportError:
                raise ImportError(
                    "Could not find the livereload library. Install it by running 'pip install livereload==2.5.1'"
                )

            import glob

            server = Server(self.app)
            for filepath in glob.glob("resources/templates/**/*/"):
                server.watch(filepath)

            self.line("")
            self.info("Live reload server is starting...")
            self.info(
                "This will only work for templates. Changes to Python files may require a browser refresh."
            )
            self.line("")
            server.serve(
                port=self.option("port"),
                restart_delay=self.option("reload-interval"),
                liveport=5500,
                root=self.app.base_path,
                debug=True,
            )
            return

        use_reloader = True
        threaded = False
        extra_files = [".env", self.app.get_storage_path()]

        if self.option("dont-reload"):
            use_reloader = False

        if self.option("threaded"):
            threaded = True

        run_simple(
            self.option("host"),
            int(self.option("port")),
            self.app,
            threaded=threaded,
            use_reloader=use_reloader,
            extra_files=extra_files,
            # reloader_interval=
            # more efficient than stat
            reloader_type="watchdog",
        )



from .. import Middleware


class ClearDumpsBetweenRequestsMiddleware(Middleware):
    """Clear dumps between each requests to avoid showing too much data."""

    def before(self, request, response):
        request.app.make("dumper").clear()
        return request

    def after(self, request, response):
        return request



class CanSchedule:
    def call(self, command):
        command_class = CommandTask(command)
        self.app.make("scheduler").add(command_class)
        return command_class

    def schedule(self, task):
        task_class = task
        self.app.make("scheduler").add(task_class)
        return task_class
